"""
Build faebryk core objects from ato DSL.
"""

import inspect
import itertools
import logging
import operator
import os
from collections import defaultdict
from collections.abc import Generator
from contextlib import contextmanager
from dataclasses import dataclass
from itertools import chain, pairwise
from pathlib import Path
from typing import (
    Any,
    Iterable,
    Sequence,
    Type,
    cast,
)

from antlr4 import ParserRuleContext
from more_itertools import last
from pint import UndefinedUnitError

import faebryk.library._F as F
import faebryk.libs.library.L as L
from atopile import address, errors
from atopile.attributes import GlobalAttributes, _has_ato_cmp_attrs, shim_map
from atopile.config import config
from atopile.datatypes import (
    FieldRef,
    KeyOptItem,
    KeyOptMap,
    ReferencePartType,
    StackList,
    TypeRef,
    is_int,
)
from atopile.parse import parser
from atopile.parser.AtoParser import AtoParser as ap
from atopile.parser.AtoParserVisitor import AtoParserVisitor
from faebryk.core.node import FieldExistsError, NodeException
from faebryk.core.parameter import (
    Arithmetic,
    ConstrainableExpression,
    GreaterOrEqual,
    Is,
    IsSubset,
    LessOrEqual,
    Max,
    Min,
    Parameter,
)
from faebryk.core.trait import Trait
from faebryk.libs.exceptions import accumulate, downgrade, iter_through_errors
from faebryk.libs.library.L import Range, Single
from faebryk.libs.picker.picker import does_not_require_picker_check
from faebryk.libs.sets.quantity_sets import Quantity_Interval, Quantity_Set
from faebryk.libs.sets.sets import BoolSet
from faebryk.libs.units import (
    HasUnit,
    P,
    Quantity,
    UnitCompatibilityError,
    dimensionless,
)
from faebryk.libs.units import (
    Unit as UnitType,
)
from faebryk.libs.util import (
    FuncDict,
    cast_assert,
    groupby,
    has_attr_or_property,
    has_instance_settable_attr,
    import_from_path,
    is_type_pair,
    not_none,
    partition_as_list,
)

logger = logging.getLogger(__name__)


Numeric = Parameter | Arithmetic | Quantity_Set


class from_dsl(Trait.decless()):
    def __init__(self, src_ctx: ParserRuleContext) -> None:
        super().__init__()
        self.src_ctx = src_ctx


class BasicsMixin:
    def visitName(self, ctx: ap.NameContext) -> str:
        """
        If this is an int, convert it to one (for pins),
        else return the name as a string.
        """
        return ctx.getText()

    def visitTypeReference(self, ctx: ap.Type_referenceContext) -> TypeRef:
        return TypeRef(self.visitName(name) for name in ctx.name())

    def visitArrayIndex(self, ctx: ap.Array_indexContext | None) -> str | int | None:
        if ctx is None:
            return None
        if key := ctx.key():
            out = key.getText()
            if is_int(out):
                return int(out)
            return out
        return None

    def visitFieldReferencePart(
        self, ctx: ap.Field_reference_partContext
    ) -> ReferencePartType:
        return ReferencePartType(
            self.visitName(ctx.name()), self.visitArrayIndex(ctx.array_index())
        )

    def visitFieldReference(self, ctx: ap.Field_referenceContext) -> FieldRef:
        pin = ctx.pin_reference_end()
        if pin is not None:
            pin = int(pin.NUMBER().getText())
        return FieldRef(
            parts=(
                self.visitFieldReferencePart(part)
                for part in ctx.field_reference_part()
            ),
            pin=pin,
        )

    def visitString(self, ctx: ap.StringContext) -> str:
        raw: str = ctx.getText()
        return raw.strip("\"'")

    def visitBoolean_(self, ctx: ap.Boolean_Context) -> bool:
        raw: str = ctx.getText()

        if raw.lower() == "true":
            return True
        elif raw.lower() == "false":
            return False

        raise errors.UserException.from_ctx(ctx, f"Expected a boolean value, got {raw}")


class NOTHING:
    """A sentinel object to represent a "nothing" return value."""


class SkipPriorFailedException(Exception):
    """Raised to skip a statement in case a dependency already failed"""


class DeprecatedException(errors.UserException):
    """
    Raised when a deprecated feature is used.
    """

    def get_frozen(self) -> tuple:
        # TODO: this is a bit of a hack to make the logger de-dup these for us
        return errors._BaseBaseUserException.get_frozen(self)


class SequenceMixin:
    """
    The base translator is responsible for methods common to
    navigating from the top of the AST including how to process
    errors, and commonising return types.
    """

    def defaultResult(self):
        """
        Override the default "None" return type
        (for things that return nothing) with the Sentinel NOTHING
        """
        return NOTHING

    def visit_iterable_helper(self, children: Iterable) -> KeyOptMap:
        """
        Visit multiple children and return a tuple of their results,
        discard any results that are NOTHING and flattening the children's results.
        It is assumed the children are returning their own OptionallyNamedItems.
        """

        def _visit():
            for err_cltr, child in iter_through_errors(
                children,
                errors._BaseBaseUserException,
                SkipPriorFailedException,
            ):
                with err_cltr():
                    # Since we're in a SequenceMixin, we need to cast self to the visitor type # noqa: E501  # pre-existing
                    child_result = cast(AtoParserVisitor, self).visit(child)
                    if child_result is not NOTHING:
                        yield child_result

        child_results = chain.from_iterable(_visit())
        child_results = filter(lambda x: x is not NOTHING, child_results)
        child_results = KeyOptMap(KeyOptItem(cr) for cr in child_results)

        return KeyOptMap(child_results)

    def visitFile_input(self, ctx: ap.File_inputContext) -> KeyOptMap:
        return self.visit_iterable_helper(ctx.stmt())

    def visitSimple_stmts(self, ctx: ap.Simple_stmtsContext) -> KeyOptMap:
        return self.visit_iterable_helper(ctx.simple_stmt())

    def visitBlock(self, ctx: ap.BlockContext) -> KeyOptMap:
        if ctx.stmt():
            return self.visit_iterable_helper(ctx.stmt())
        if ctx.simple_stmts():
            return self.visitSimple_stmts(ctx.simple_stmts())

        raise ValueError  # this should be protected because it shouldn't be parseable


class BlockNotFoundError(errors.UserKeyError):
    """
    Raised when a block doesn't exist.
    """


@dataclass
class Context:
    """~A metaclass to hold context/origin information on ato classes."""

    @dataclass
    class ImportPlaceholder:
        ref: TypeRef
        from_path: str
        original_ctx: ParserRuleContext

    # Location information re. the source of this module
    file_path: Path | None

    # Scope information
    scope_ctx: ap.BlockdefContext | ap.File_inputContext
    refs: dict[TypeRef, Type[L.Node] | ap.BlockdefContext | ImportPlaceholder]


class Wendy(BasicsMixin, SequenceMixin, AtoParserVisitor):  # type: ignore  # Overriding base class makes sense here
    """
    Wendy is Bob's business partner and fellow builder in the children's TV series
    "Bob the Builder." She is a skilled construction worker who often manages the
    business side of their building company while also participating in hands-on
    construction work. Wendy is portrayed as capable, practical and level-headed,
    often helping to keep projects organized and on track. She wears a green safety
    helmet and work clothes, and is known for her competence in operating various
    construction vehicles and equipment.

    Wendy also knows where to find the best building supplies.
    """

    def visitImport_stmt(
        self, ctx: ap.Import_stmtContext
    ) -> KeyOptMap[tuple[Context.ImportPlaceholder, ap.Import_stmtContext]]:
        if from_path := ctx.string():
            lazy_imports = [
                Context.ImportPlaceholder(
                    ref=self.visitTypeReference(reference),
                    from_path=self.visitString(from_path),
                    original_ctx=ctx,
                )
                for reference in ctx.type_reference()
            ]
            return KeyOptMap(
                KeyOptItem.from_kv(li.ref, (li, ctx)) for li in lazy_imports
            )

        else:
            # Standard library imports are special, and don't require a from path
            imports = []
            for collector, reference in iter_through_errors(ctx.type_reference()):
                with collector():
                    ref = self.visitTypeReference(reference)
                    if len(ref) > 1:
                        raise errors.UserKeyError.from_ctx(
                            ctx, "Standard library imports must be single-name"
                        )

                    name = ref[0]
                    if not hasattr(F, name) or not issubclass(
                        getattr(F, name), (L.Module, L.ModuleInterface)
                    ):
                        raise errors.UserKeyError.from_ctx(
                            ctx, f"Unknown standard library module: '{name}'"
                        )

                    imports.append(KeyOptItem.from_kv(ref, (getattr(F, name), ctx)))

            return KeyOptMap(imports)

    def visitDep_import_stmt(
        self, ctx: ap.Dep_import_stmtContext
    ) -> KeyOptMap[tuple[Context.ImportPlaceholder, ap.Dep_import_stmtContext]]:
        lazy_import = Context.ImportPlaceholder(
            ref=self.visitTypeReference(ctx.type_reference()),
            from_path=self.visitString(ctx.string()),
            original_ctx=ctx,
        )
        # TODO: @v0.4 remove this deprecated import form
        with downgrade(DeprecatedException):
            raise DeprecatedException.from_ctx(
                ctx,
                "`import <something> from <path>` is deprecated and"
                " will be removed in a future version. Use "
                f"`from {ctx.string().getText()} import"
                f" {ctx.type_reference().getText()}`"
                " instead.",
            )
        return KeyOptMap.from_kv(lazy_import.ref, (lazy_import, ctx))

    def visitBlockdef(
        self, ctx: ap.BlockdefContext
    ) -> KeyOptMap[tuple[ap.BlockdefContext, ap.BlockdefContext]]:
        ref = TypeRef.from_one(self.visitName(ctx.name()))
        return KeyOptMap.from_kv(ref, (ctx, ctx))

    def visitSimple_stmt(
        self, ctx: ap.Simple_stmtContext | Any
    ) -> (
        KeyOptMap[
            tuple[Context.ImportPlaceholder | ap.BlockdefContext, ParserRuleContext]
        ]
        | type[NOTHING]
    ):
        if ctx.import_stmt() or ctx.dep_import_stmt():
            return super().visitChildren(ctx)
        return NOTHING

    # TODO: @v0.4: remove this shimming
    @staticmethod
    def _find_shim(
        file_path: Path | None, ref: TypeRef
    ) -> tuple[Type[L.Node], str] | None:
        if file_path is None:
            return None

        import_addr = address.AddrStr.from_parts(file_path, str(TypeRef(ref)))

        for shim_addr in shim_map:
            if import_addr.endswith(shim_addr):
                return shim_map[shim_addr]

        return None

    @classmethod
    def survey(
        cls, file_path: Path | None, ctx: ap.BlockdefContext | ap.File_inputContext
    ) -> Context:
        surveyor = cls()
        context = Context(file_path=file_path, scope_ctx=ctx, refs={})
        for ref, (item, item_ctx) in surveyor.visit(ctx):
            assert isinstance(item_ctx, ParserRuleContext)
            if ref in context.refs:
                # Downgrade the error in case we're shadowing things
                # Not limiting the number of times we show this warning
                # because they're pretty important and Wendy is well cached
                with downgrade(errors.UserKeyError):
                    raise errors.UserKeyError.from_ctx(
                        item_ctx,
                        f"`{ref}` already declared. Shadowing original."
                        " In the future this may be an error",
                    )

            # TODO: @v0.4: remove this shimming
            if shim := cls._find_shim(context.file_path, ref):
                shim_cls, preferred = shim

                if hasattr(item_ctx, "name"):
                    dep_ctx = item_ctx.name()  # type: ignore
                elif hasattr(item_ctx, "reference"):
                    dep_ctx = item_ctx.reference()  # type: ignore
                else:
                    dep_ctx = item_ctx

                # TODO: @v0.4 increase the level of this to WARNING
                # when there's an alternative
                with downgrade(DeprecatedException, to_level=logging.DEBUG):
                    raise DeprecatedException.from_ctx(
                        dep_ctx,
                        f"`{ref}` is deprecated and will be removed in a future"
                        f" version. Use `{preferred}` instead.",
                    )

                context.refs[ref] = shim_cls
            else:
                context.refs[ref] = item

        return context


@contextmanager
def ato_error_converter():
    try:
        yield
    except NodeException as ex:
        if from_dsl_ := ex.node.try_get_trait(from_dsl):
            raise errors.UserException.from_ctx(from_dsl_.src_ctx, str(ex)) from ex
        else:
            raise ex


@contextmanager
def _attach_ctx_to_ex(ctx: ParserRuleContext, traceback: Sequence[ParserRuleContext]):
    try:
        yield
    except errors.UserException as ex:
        if ex.origin_start is None:
            ex.attach_origin_from_ctx(ctx)
            # only attach traceback if we're also setting the origin
            if ex.traceback is None:
                ex.traceback = traceback
        raise ex


_declaration_domain_to_unit = {
    "dimensionless": dimensionless,
    "resistance": P.ohm,
    "capacitance": P.farad,
    "inductance": P.henry,
    "voltage": P.volt,
    "current": P.ampere,
    "power": P.watt,
    "frequency": P.hertz,
}


@dataclass
class _ParameterDefinition:
    """
    Holds information about a parameter declaration or assignment.
    We collect those per parameter in the `Bob._param_assignments` dict.
    Multiple assignments are allowed, but they interact with each other in non-trivial
    ways. Thus we need to track them and process them in the end in
    `_merge_parameter_assignments`.
    """

    ctx: ParserRuleContext
    traceback: Sequence[ParserRuleContext]
    ref: FieldRef
    value: Range | Single | None = None

    @property
    def is_declaration(self) -> bool:
        return self.value is None

    @property
    def is_definition(self) -> bool:
        return not self.is_declaration

    def __post_init__(self):
        pass

    @property
    def is_root_assignment(self) -> bool:
        if not self.is_definition:
            return False
        return len(self.ref) == 1


class Bob(BasicsMixin, SequenceMixin, AtoParserVisitor):  # type: ignore  # Overriding base class makes sense here
    """
    Bob is a general contractor who runs his own construction company in the town
    of Fixham Harbour (in earlier episodes, he was based in Bobsville). Recognizable
    by his blue jeans, checked shirt, yellow hard hat, and tool belt, Bob is known
    for his positive catchphrase "Can we fix it? Yes, we can!" He's portrayed as a
    friendly, optimistic problem-solver who takes pride in helping his community
    through various building and repair projects. Bob works closely with his team
    of anthropomorphic construction vehicles and his business partner Wendy,
    tackling each construction challenge with enthusiasm and determination. His
    character embodies values of teamwork, perseverance, and taking pride in one's
    work.
    """

    def __init__(self) -> None:
        super().__init__()
        self._scopes = FuncDict[ParserRuleContext, Context]()
        self._python_classes = FuncDict[ap.BlockdefContext, Type[L.Module]]()
        self._node_stack = StackList[L.Node]()
        self._traceback_stack = StackList[ParserRuleContext]()

        self._param_assignments = defaultdict[Parameter, list[_ParameterDefinition]](
            list
        )
        self.search_paths: list[os.PathLike] = []

        # Keeps track of the nodes whose construction failed,
        # so we don't report dud key errors when it was a higher failure
        # that caused the node not to exist
        self._failed_nodes = FuncDict[L.Node, set[str]]()

    def build_ast(
        self, ast: ap.File_inputContext, ref: TypeRef, file_path: Path | None = None
    ) -> L.Node:
        """Build a Module from an AST and reference."""
        file_path = self._sanitise_path(file_path) if file_path else None
        context = self.index_ast(ast, file_path)
        return self._build(context, ref)

    def build_file(self, path: Path, ref: TypeRef) -> L.Node:
        """Build a Module from a file and reference."""
        context = self.index_file(self._sanitise_path(path))
        return self._build(context, ref)

    @property
    def modules(self) -> dict[address.AddrStr, Type[L.Module]]:
        """Conceptually similar to `sys.modules`"""

        # FIXME: this feels like a shit way to get addresses of the imported modules
        def _get_addr(ctx: ParserRuleContext):
            ref = tuple()
            ctx_ = ctx
            while ctx_ not in self._scopes:
                if isinstance(ctx_, ap.BlockdefContext):
                    ref = (ctx_.name().getText(),) + ref
                ctx_ = ctx_.parentCtx
                if ctx_ is None:
                    return None

            return address.AddrStr.from_parts(
                self._scopes[ctx_].file_path, str(TypeRef(ref))
            )

        return {
            addr: cls
            for ctx, cls in self._python_classes.items()
            if (addr := _get_addr(ctx)) is not None
        }

    def _build(self, context: Context, ref: TypeRef) -> L.Node:
        assert self._is_reset()

        if ref not in context.refs:
            raise errors.UserKeyError.from_ctx(
                context.scope_ctx, f"No declaration of `{ref}` in {context.file_path}"
            )
        try:
            class_ = self._get_referenced_class(context.scope_ctx, ref)
            if not isinstance(class_, ap.BlockdefContext):
                raise errors.UserNotImplementedError(
                    "Can't initialize a fabll directly like this"
                )
            with self._traceback_stack.enter(class_.name()):
                with self._init_node(class_) as node:
                    node.add(F.is_app_root())
                return node
        except* SkipPriorFailedException:
            raise errors.UserException("Build failed")
        finally:
            self._finish()

    def _is_reset(self) -> bool:
        """
        Make sure caches that aren't intended to be shared between builds are empty.
        True if the caches are empty, False if they are not.
        """
        return (
            not self._node_stack
            and not self._traceback_stack
            and not self._param_assignments
        )

    def _finish(self):
        self._merge_parameter_assignments()
        assert self._is_reset()

    class ParamAssignmentIsGospel(errors.UserException):
        """
        The parameter assignment is treated as the precise specification of the
        component, rather than merely as a requirement for it's later selection
        """

        title = "Parameter assignments are component definition"  # type: ignore

    def _merge_parameter_assignments(self):
        with accumulate(
            errors._BaseBaseUserException, SkipPriorFailedException
        ) as ex_acc:
            # Handle missing definitions
            params_without_defitions, params_with_definitions = partition_as_list(
                lambda p: any(a.is_definition for a in self._param_assignments[p]),
                self._param_assignments,
            )

            for param in params_without_defitions:
                last_declaration = last(self._param_assignments.pop(param))
                with ex_acc.collect(), ato_error_converter():
                    # TODO: @v0.4 remove this deprecated import form
                    with downgrade(
                        errors.UserActionWithoutEffectError, to_level=logging.DEBUG
                    ):
                        raise errors.UserActionWithoutEffectError.from_ctx(
                            last_declaration.ctx,
                            f"Attribute `{param}` declared but never assigned.",
                            traceback=last_declaration.traceback,
                        )
            # Handle parameter assignments
            # assignments override each other
            # assignments made in the block definition of the component are "is"
            # external assignments are requirements treated as a subset

            # Allowing external assignments in the first place is a bit weird
            # Got to figure out how people are using this.
            # My guess is that in 99% of cases you can replace them by a `&=`
            params_by_node = groupby(
                params_with_definitions, key=lambda p: p.get_parent_force()[0]
            )
            for assignee_node, assigned_params in params_by_node.items():
                is_part_module = isinstance(assignee_node, L.Module) and (
                    assignee_node.has_trait(F.is_pickable_by_supplier_id)
                    or assignee_node.has_trait(F.is_pickable_by_part_number)
                )

                gospel_params: list[Parameter] = []
                for param in assigned_params:
                    assignments = self._param_assignments.pop(param)
                    definitions = [a for a in assignments if a.is_definition]
                    non_root_definitions, root_definitions = partition_as_list(
                        lambda a: a.is_root_assignment, definitions
                    )

                    assert definitions
                    for definition in definitions:
                        logger.debug(
                            "Assignment:  %s [%s] := %s",
                            param,
                            definition.ref,
                            definition.value,
                        )

                    # Don't see how this could happen, but just in case
                    root_after_external = (False, True) in pairwise(
                        a.is_root_assignment for a in definitions
                    )
                    assert not root_after_external

                    with ex_acc.collect(), ato_error_converter():
                        # Workaround for missing difference between alias and subset
                        # Only relevant for code-as-data part modules
                        # TODO: consider a warning for definitions that aren't purely
                        # narrowing
                        if is_part_module and non_root_definitions:
                            raise errors.UserNotImplementedError.from_ctx(
                                last(non_root_definitions).ctx,
                                "You can't assign to a `component` with a specific"
                                " part number outside of its definition",
                                traceback=last(non_root_definitions).traceback,
                            )

                        elif is_part_module and root_definitions:
                            param.alias_is(not_none(last(root_definitions).value))
                            param.add(does_not_require_picker_check())
                            gospel_params.append(param)

                        elif not is_part_module:
                            definition = last(definitions)
                            value = not_none(definition.value)
                            try:
                                logger.debug("Constraining %s to %s", param, value)
                                param.constrain_subset(value)
                            except UnitCompatibilityError as ex:
                                raise errors.UserTypeError.from_ctx(
                                    definition.ctx,
                                    str(ex),
                                    traceback=definition.traceback,
                                ) from ex

                if gospel_params:
                    with downgrade(self.ParamAssignmentIsGospel, to_level=logging.INFO):
                        raise self.ParamAssignmentIsGospel(
                            f"`component` `{assignee_node.get_full_name()}`"
                            " is completely specified by a part number, so these"
                            " params are treated as its exact specification:"
                            + ", ".join(f"`{p}`" for p in gospel_params)
                        )

    @property
    def _current_node(self) -> L.Node:
        return self._node_stack[-1]

    def get_traceback(self) -> Sequence[ParserRuleContext]:
        """Return the current traceback, with sequential duplicates removed"""
        # Use dict ordering guarantees and key uniqueness to remove duplicates
        return list(dict.fromkeys(self._traceback_stack).keys())

    @staticmethod
    def _sanitise_path(path: os.PathLike) -> Path:
        return Path(path).expanduser().resolve().absolute()

    def index_ast(
        self, ast: ap.File_inputContext, file_path: Path | None = None
    ) -> Context:
        if ast in self._scopes:
            return self._scopes[ast]

        context = Wendy.survey(file_path, ast)
        self._scopes[ast] = context
        return context

    def index_file(self, file_path: Path) -> Context:
        ast = parser.get_ast_from_file(file_path)
        return self.index_ast(ast, file_path)

    def _get_search_paths(self, context: Context) -> list[Path]:
        search_paths = [Path(p) for p in self.search_paths]

        if context.file_path is not None:
            search_paths.insert(0, context.file_path.parent)

        if config.has_project:
            search_paths += [config.project.paths.src, config.project.paths.modules]

        # Add the library directory to the search path too
        search_paths.append(Path(inspect.getfile(F)).parent)

        return search_paths

    def _import_item(
        self, context: Context, item: Context.ImportPlaceholder
    ) -> Type[L.Node] | ap.BlockdefContext:
        # Build up search paths to check for the import in
        # Iterate though them, checking if any contains the thing we're looking for
        search_paths = self._get_search_paths(context)
        for search_path in search_paths:
            candidate_from_path = search_path / item.from_path
            if candidate_from_path.exists():
                break
        else:
            raise errors.UserFileNotFoundError.from_ctx(
                item.original_ctx,
                f"Can't find {item.from_path} in {', '.join(map(str, search_paths))}",
            )

        from_path = self._sanitise_path(candidate_from_path)
        if from_path.suffix == ".py":
            try:
                node = import_from_path(from_path)
            except FileNotFoundError as ex:
                raise errors.UserImportNotFoundError.from_ctx(
                    item.original_ctx, str(ex)
                ) from ex

            for ref in item.ref:
                try:
                    node = getattr(node, ref)
                except AttributeError as ex:
                    raise errors.UserKeyError.from_ctx(
                        item.original_ctx, f"No attribute `{ref}` found on {node}"
                    ) from ex

            assert isinstance(node, type) and issubclass(node, L.Node)
            return node

        elif from_path.suffix == ".ato":
            context = self.index_file(from_path)
            if item.ref not in context.refs:
                raise errors.UserKeyError.from_ctx(
                    item.original_ctx, f"No declaration of `{item.ref}` in {from_path}"
                )
            node = context.refs[item.ref]

            if isinstance(node, Context.ImportPlaceholder):
                raise errors.UserTypeError.from_ctx(
                    item.original_ctx,
                    "Importing a import is not supported",
                )

            assert (
                isinstance(node, type)
                and issubclass(node, L.Node)
                or isinstance(node, ap.BlockdefContext | ap.File_inputContext)
            )
            return node

        else:
            raise errors.UserImportNotFoundError.from_ctx(
                item.original_ctx, f"Can't import file type {from_path.suffix}"
            )

    def _get_referenced_class(
        self, ctx: ParserRuleContext, ref: TypeRef
    ) -> Type[L.Node] | ap.BlockdefContext:
        """
        Returns the class / object referenced by the given ref,
        based on Bob's current context. The contextual nature
        of this means that it's only useful during the build process.
        """
        # No change in position from the current context
        # return self, eg the current parser context
        if ref == tuple():
            if isinstance(ctx, ap.BlockdefContext):
                return ctx
            else:
                raise ValueError(f"Can't get class `{ref}` from {ctx}")

        # Ascend the tree until we find a scope that has the ref within it
        ctx_ = ctx
        while ctx_ not in self._scopes:
            if ctx_.parentCtx is None:
                raise ValueError(f"No scope found for `{ref}`")
            ctx_ = ctx_.parentCtx

        context = self._scopes[ctx_]

        # FIXME: there are more cases to check here,
        # eg. if we have part of a ref resolved
        if ref not in context.refs:
            raise errors.UserKeyError.from_ctx(
                ctx, f"No class or block definition found for `{ref}`"
            )

        item = context.refs[ref]
        # Ensure the item is resolved, if not already
        if isinstance(item, Context.ImportPlaceholder):
            # TODO: search path for these imports
            item = self._import_item(context, item)
            context.refs[ref] = item

        return item

    @staticmethod
    def get_node_attr(node: L.Node, ref: ReferencePartType) -> L.Node:
        """
        Analogous to `getattr`

        Returns the value if it exists, otherwise raises an AttributeError
        Required because we're seeing attributes in both the attrs and runtime
        """

        if has_attr_or_property(node, ref.name):
            # Build-time attributes are attached as real attributes
            result = getattr(node, ref.name)
            if ref.key is not None and isinstance(result, L.Node):
                raise ValueError(f"{ref.name} is not subscriptable")
            if not isinstance(result, L.Node) and ref.key is None:
                raise ValueError(
                    f"{ref.name} is a {type(result)._ref.name__} and needs a ref.key"
                )
            if isinstance(result, dict):
                assert ref.key is not None
                if ref.key not in result:
                    raise AttributeError(name=f"{ref.name}[{ref.key}]", obj=node)
                result = result[ref.key]
            elif isinstance(result, list):
                assert ref.key is not None
                # TODO type check key
                if not isinstance(ref.key, int):
                    raise ValueError(f"Key `{ref.key}` is not an integer")
                if ref.key >= len(result):
                    raise AttributeError(name=f"{ref.name}[{ref.key}]", obj=node)
                result = result[ref.key]
            # TODO handle non-module & non-dict & non-list case
        elif ref.name in node.runtime and ref.key is None:
            # Runtime attributes are attached as runtime attributes
            result = node.runtime[ref.name]
        else:
            # Wah wah wah - we don't know what this is
            friendlyname = ref.name if ref.key is None else f"{ref.name}[{ref.key}]"
            raise AttributeError(name=friendlyname, obj=node)

        if isinstance(result, L.Module):
            return result.get_most_special()

        return result

    def _get_referenced_node(self, ref: FieldRef, ctx: ParserRuleContext) -> L.Node:
        node = self._current_node
        for i, name in enumerate(ref):
            try:
                node = self.get_node_attr(node, name)
            except AttributeError as ex:
                # If we know that a previous failure prevented the creation
                # of this node, raise a SkipPriorFailedException to prevent
                # error messages about it missing from polluting the output
                if name in self._failed_nodes.get(node, set()):
                    raise SkipPriorFailedException() from ex

                # Wah wah wah - we don't know what this is
                # Build a nice error message
                if i > 0:
                    msg = f"`{FieldRef(ref.parts[:i])}` has no attribute `{ex.name}`"
                else:
                    msg = f"No attribute `{name}`"
                raise errors.UserKeyError.from_ctx(
                    ctx, msg, traceback=self.get_traceback()
                ) from ex
            except ValueError as ex:
                raise errors.UserKeyError.from_ctx(
                    ctx, str(ex), traceback=self.get_traceback()
                ) from ex

        return node

    def _try_get_referenced_node(
        self, ref: FieldRef, ctx: ParserRuleContext
    ) -> L.Node | None:
        try:
            return self._get_referenced_node(ref, ctx)
        except errors.UserKeyError:
            return None

    def _new_node(
        self,
        item: ap.BlockdefContext | Type[L.Node],
        promised_supers: list[ap.BlockdefContext],
    ) -> tuple[L.Node, list[ap.BlockdefContext]]:
        """
        Kind of analogous to __new__ in Python, except that it's a factory

        Descends down the class hierarchy until it finds a known base class.
        As it descends, it logs all superclasses it encounters, as `promised_supers`.
        These are accumulated lowest (base-class) to highest (what was initialised).

        Once a base class is found, it creates a new class for each superclass that
        isn't already known, attaching the __atopile_src_ctx__ attribute to the new
        class.
        """
        if isinstance(item, type) and issubclass(item, L.Node):
            super_class = item
            for super_ctx in promised_supers:
                if super_ctx in self._python_classes:
                    super_class = self._python_classes[super_ctx]
                    continue

                assert issubclass(super_class, L.Node)

                # Create a new type with a more descriptive name
                type_name = super_ctx.name().getText()
                type_qualname = f"{super_class.__module__}.{type_name}"

                super_class = type(
                    type_name,  # Class name
                    (super_class,),  # Base classes
                    {
                        "__module__": super_class.__module__,
                        "__qualname__": type_qualname,
                        "__atopile_src_ctx__": super_ctx,
                    },
                )

                self._python_classes[super_ctx] = super_class

            assert issubclass(super_class, L.Node)
            return super_class(), promised_supers

        if isinstance(item, ap.BlockdefContext):
            # Find the superclass of the new node, if there's one defined
            block_type = item.blocktype()
            if super_ctx := item.blockdef_super():
                super_ref = self.visitTypeReference(super_ctx.type_reference())
                # Create a base node to build off
                base_class = self._get_referenced_class(item, super_ref)
            else:
                # Create a shell of base-node to build off
                assert isinstance(block_type, ap.BlocktypeContext)
                if block_type.INTERFACE():
                    base_class = L.ModuleInterface
                elif block_type.COMPONENT():
                    base_class = L.Module
                elif block_type.MODULE():
                    base_class = L.Module
                else:
                    raise ValueError(f"Unknown block type `{block_type.getText()}`")

            # Descend into building the superclass. We've got no information
            # on when the super-chain will be resolved, so we need to promise
            # that this current blockdef will be visited as part of the init
            result = self._new_node(
                base_class,
                promised_supers=[item] + promised_supers,
            )

            return result

        # This should never happen
        raise ValueError(f"Unknown item type `{item}`")

    @contextmanager
    def _init_node(
        self, node_type: ap.BlockdefContext | Type[L.Node]
    ) -> Generator[L.Node, None, None]:
        """
        Kind of analogous to __init__ in Python, except that it's a factory

        Pre-yield it is analogous to __new__, where it creates the hollow instance
        Post-yield it is analogous to __init__, where it fills in the details

        This is to allow for it to be attached in the graph before it's filled,
        and subsequently for errors to be raised in context of it's graph location.
        """
        new_node, promised_supers = self._new_node(
            node_type,
            promised_supers=[],
        )

        # Shim on component and module classes defined in ato
        # Do not shim fabll modules, or interfaces
        if isinstance(node_type, ap.BlockdefContext):
            if node_type.blocktype().COMPONENT() or node_type.blocktype().MODULE():
                # Some shims add the trait themselves
                if not new_node.has_trait(_has_ato_cmp_attrs):
                    new_node.add(_has_ato_cmp_attrs())

        yield new_node

        with self._node_stack.enter(new_node):
            for super_ctx in promised_supers:
                # TODO: this would be better if we had the
                # "from xyz" super in the traceback too
                with self._traceback_stack.enter(super_ctx.name()):
                    self.visitBlock(super_ctx.block())

    def _get_param(
        self, node: L.Node, ref: ReferencePartType, src_ctx: ParserRuleContext
    ) -> Parameter:
        """
        Get a param from a node.
        Not supported: If it doesn't exist, create it and promise to assign
        it later. Used in forward-declaration.
        """
        try:
            node = self.get_node_attr(node, ref)
        except AttributeError as ex:
            if ref in self._failed_nodes.get(node, set()):
                raise SkipPriorFailedException() from ex
            # Wah wah wah - we don't know what this is
            raise errors.UserNotImplementedError.from_ctx(
                src_ctx,
                f"Parameter `{ref}` not found and"
                " forward-declared params are not yet implemented",
                traceback=self.get_traceback(),
            ) from ex
        except ValueError as ex:
            raise errors.UserValueError.from_ctx(
                src_ctx, str(ex), traceback=self.get_traceback()
            ) from ex

        if not isinstance(node, Parameter):
            raise errors.UserSyntaxError.from_ctx(
                src_ctx,
                f"Node {ref} is {type(node)} not a Parameter",
                traceback=self.get_traceback(),
            )
        return node

    def _ensure_param(
        self,
        node: L.Node,
        ref: ReferencePartType,
        unit: UnitType,
        src_ctx: ParserRuleContext,
    ) -> Parameter:
        """
        Ensure a node has a param with a given name
        If it already exists, check the unit is compatible and return it
        """

        try:
            param = self.get_node_attr(node, ref)
        except AttributeError:
            # Here we attach only minimal information, so we can override it later
            if ref.key is not None:
                if not isinstance(ref.key, str):
                    raise errors.UserNotImplementedError.from_ctx(
                        src_ctx,
                        f"Can't forward assign to a non-string key `{ref}`",
                        traceback=self.get_traceback(),
                    )
                container = getattr(node, ref.name)
                param = node.add(
                    Parameter(units=unit, domain=L.Domains.Numbers.REAL()),
                    name=ref.key,
                    container=container,
                )
            else:
                param = node.add(
                    Parameter(units=unit, domain=L.Domains.Numbers.REAL()),
                    name=ref.name,
                )
        except ValueError as ex:
            raise errors.UserValueError.from_ctx(
                src_ctx, str(ex), traceback=self.get_traceback()
            ) from ex
        else:
            if not isinstance(param, Parameter):
                raise errors.UserTypeError.from_ctx(
                    src_ctx,
                    f"Cannot assign a parameter to `{ref}` on `{node}` because its"
                    f" type is `{param.__class__.__name__}`",
                    traceback=self.get_traceback(),
                )

        if not param.units.is_compatible_with(unit):
            raise errors.UserIncompatibleUnitError.from_ctx(
                src_ctx,
                f"Given units ({unit}) are incompatible"
                f" with existing units ({param.units}).",
                traceback=self.get_traceback(),
            )

        return param

    def _record_failed_node(self, node: L.Node, name: str):
        self._failed_nodes.setdefault(node, set()).add(name)

    def visitAssign_stmt(self, ctx: ap.Assign_stmtContext):
        """Assignment values and create new instance of things."""
        dec = ctx.field_reference_or_declaration()
        assigned_ref = self.visitFieldReference(
            dec.field_reference() or dec.declaration_stmt().field_reference()
        )

        assigned_name: ReferencePartType = assigned_ref[-1]
        assignable_ctx = ctx.assignable()
        assert isinstance(assignable_ctx, ap.AssignableContext)
        target = self._get_referenced_node(assigned_ref.stem, ctx)

        ########## Handle New Statements ##########
        if new_stmt_ctx := assignable_ctx.new_stmt():
            if len(assigned_ref) > 1:
                raise errors.UserSyntaxError.from_ctx(
                    ctx,
                    f"Can't declare fields in a nested object `{assigned_ref}`",
                    traceback=self.get_traceback(),
                )
            if assigned_name.key is not None:
                raise errors.UserSyntaxError.from_ctx(
                    ctx,
                    f"Can't use keys with `new` statements `{assigned_ref}`",
                    traceback=self.get_traceback(),
                )

            assert isinstance(new_stmt_ctx, ap.New_stmtContext)
            ref = self.visitTypeReference(new_stmt_ctx.type_reference())

            try:
                with self._traceback_stack.enter(new_stmt_ctx):
                    with self._init_node(
                        self._get_referenced_class(ctx, ref)
                    ) as new_node:
                        try:
                            self._current_node.add(new_node, name=assigned_name.name)
                        except FieldExistsError as e:
                            raise errors.UserAlreadyExistsError.from_ctx(
                                ctx,
                                f"Field `{assigned_name}` already exists",
                                traceback=self.get_traceback(),
                            ) from e
                        new_node.add(from_dsl(ctx))
            except Exception:
                # Not a narrower exception because it's often an ExceptionGroup
                self._record_failed_node(self._current_node, assigned_name.name)
                raise

            return NOTHING

        ########## Handle Regular Assignments ##########
        value = self.visit(assignable_ctx)
        # Arithmetic
        if assignable_ctx.literal_physical() or assignable_ctx.arithmetic_expression():
            declaration = ctx.field_reference_or_declaration().declaration_stmt()
            if declaration:
                # check valid declaration
                # create param with corresponding units
                self.visitDeclaration_stmt(declaration)

            unit = HasUnit.get_units(value)
            param = self._ensure_param(target, assigned_name, unit, ctx)
            self._param_assignments[param].append(
                _ParameterDefinition(
                    ref=assigned_ref,
                    value=value,
                    ctx=ctx,
                    traceback=self.get_traceback(),
                )
            )

        # String or boolean
        elif assignable_ctx.string() or assignable_ctx.boolean_():
            if assigned_name.key is not None:
                raise errors.UserSyntaxError.from_ctx(
                    ctx,
                    f"Can't use keys with non-arithmetic attribute assignments "
                    f"`{assigned_ref}`",
                    traceback=self.get_traceback(),
                )

            # Check if it's a property or attribute that can be set
            if has_instance_settable_attr(target, assigned_name.name):
                try:
                    setattr(target, assigned_name.name, value)
                except errors.UserException as e:
                    e.attach_origin_from_ctx(assignable_ctx)
                    raise
            elif (
                # If ModuleShims has a settable property, use it
                hasattr(GlobalAttributes, assigned_name.name)
                and isinstance(getattr(GlobalAttributes, assigned_name.name), property)
                and getattr(GlobalAttributes, assigned_name.name).fset
            ):
                prop = cast_assert(
                    property, getattr(GlobalAttributes, assigned_name.name)
                )
                assert prop.fset is not None
                # TODO: @v0.4 remove this deprecated import form
                with (
                    downgrade(DeprecatedException, errors.UserNotImplementedError),
                    _attach_ctx_to_ex(ctx, self.get_traceback()),
                ):
                    prop.fset(target, value)
            else:
                # Strictly, these are two classes of errors that could use independent
                # suppression, but we'll just suppress them both collectively for now
                # TODO: @v0.4 remove this deprecated import form
                with downgrade(errors.UserException):
                    raise errors.UserException.from_ctx(
                        ctx,
                        f"Ignoring assignment of `{value}` to `{assigned_name}` on"
                        f" `{target}`",
                        traceback=self.get_traceback(),
                    )

        else:
            raise ValueError(f"Unhandled assignable type `{assignable_ctx.getText()}`")

        return NOTHING

    def _get_mif_and_warn_when_exists(
        self, name: ReferencePartType, ctx: ParserRuleContext
    ) -> L.ModuleInterface | None:
        try:
            mif = self.get_node_attr(self._current_node, name)
        except AttributeError:
            return None
        except ValueError as ex:
            raise errors.UserValueError.from_ctx(
                ctx, str(ex), traceback=self.get_traceback()
            ) from ex

        if isinstance(mif, L.ModuleInterface):
            # TODO: @v0.4 remove this deprecated import form
            with downgrade(errors.UserAlreadyExistsError):
                raise errors.UserAlreadyExistsError(
                    f"`{name}` already exists; skipping."
                )
        else:
            raise errors.UserTypeError.from_ctx(
                ctx,
                f"`{name}` already exists.",
                traceback=self.get_traceback(),
            )

        return mif

    def visitPindef_stmt(
        self, ctx: ap.Pindef_stmtContext
    ) -> KeyOptMap[L.ModuleInterface]:
        return self.visitPin_stmt(ctx.pin_stmt(), declaration=False)

    def visitPin_declaration(
        self, ctx: ap.Pin_declarationContext
    ) -> KeyOptMap[L.ModuleInterface]:
        return self.visitPin_stmt(ctx.pin_stmt(), declaration=True)

    def visitPin_stmt(
        self, ctx: ap.Pin_stmtContext, declaration: bool
    ) -> KeyOptMap[L.ModuleInterface]:
        if ctx.name():
            name = self.visitName(ctx.name())
        elif ctx.totally_an_integer():
            name = f"{ctx.totally_an_integer().getText()}"
        elif ctx.string():
            name = self.visitString(ctx.string())
        else:
            raise ValueError(f"Unhandled pin name type `{ctx}`")

        ref = FieldRef(parts=[], pin=name).last
        if declaration:
            if mif := self._get_mif_and_warn_when_exists(ref, ctx):
                return KeyOptMap.from_item(
                    KeyOptItem.from_kv(TypeRef.from_one(name), mif)
                )
        else:
            try:
                mif = self.get_node_attr(self._current_node, ref)
            except AttributeError:
                pass
            else:
                return KeyOptMap.from_item(
                    KeyOptItem.from_kv(TypeRef.from_one(name), mif)
                )

        if shims_t := self._current_node.try_get_trait(_has_ato_cmp_attrs):
            mif = shims_t.add_pin(name, ref.name)
            return KeyOptMap.from_item(KeyOptItem.from_kv(TypeRef.from_one(name), mif))

        raise errors.UserTypeError.from_ctx(
            ctx,
            f"Can't declare pins on components of type {self._current_node}",
            traceback=self.get_traceback(),
        )

    def visitSignaldef_stmt(
        self, ctx: ap.Signaldef_stmtContext
    ) -> KeyOptMap[L.ModuleInterface]:
        name = self.visitName(ctx.name())
        # TODO: @v0.4: remove this protection
        if mif := self._get_mif_and_warn_when_exists(ReferencePartType(name), ctx):
            return KeyOptMap.from_item(KeyOptItem.from_kv(TypeRef.from_one(name), mif))

        mif = self._current_node.add(F.Electrical(), name=name)
        return KeyOptMap.from_item(KeyOptItem.from_kv(TypeRef.from_one(name), mif))

    def _connect(
        self, a: L.ModuleInterface, b: L.ModuleInterface, ctx: ParserRuleContext | None
    ):
        """
        FIXME: In ato, we allowed duck-typing of connectables
        We need to reconcile this with the strong typing
        in faebryk's connect method
        For now, we'll attempt to connect by name, and log a deprecation
        warning if that succeeds, else, re-raise the exception emitted
        by the connect method
        """
        # If we're attempting to connect an Electrical to a SignalElectrical
        # (or ElectricLogic) then allow the connection, but issue a warning
        if pair := is_type_pair(a, b, F.Electrical, F.ElectricSignal):
            pair[0].connect(pair[1].line)

            # TODO: @v0.4 remove this deprecated import form
            with downgrade(errors.UserTypeError):
                a, b = pair
                a_type = a.__class__.__name__
                b_type = b.__class__.__name__
                raise errors.UserTypeError.from_ctx(
                    ctx,
                    f"Connected `{a}` (type {a_type}) to "
                    f"`{b}.line`, because `{b}` is an `{b_type}`. "
                    "This means that the reference isn't also connected through.",
                    traceback=self.get_traceback(),
                )

        else:
            try:
                # Try a proper connection
                a.connect(b)

            except NodeException as top_ex:
                top_ex = errors.UserNodeException.from_node_exception(
                    top_ex, ctx, self.get_traceback()
                )

                # If that fails, try connecting via duck-typing
                for name, (c_a, c_b) in a.zip_children_by_name_with(
                    b, L.ModuleInterface
                ).items():
                    if c_a is None:
                        if has_attr_or_property(a, name):
                            c_a = getattr(a, name)
                        else:
                            raise top_ex

                    if c_b is None:
                        if has_attr_or_property(b, name):
                            c_b = getattr(b, name)
                        else:
                            raise top_ex

                    try:
                        self._connect(c_a, c_b, None)
                    except NodeException:
                        raise top_ex

                else:
                    # If we connect everything via name (and tried in the first place)
                    # then we're good to go! We just need to tell everyone to probably
                    # not do that in the future - and we're off!
                    if (
                        ctx is not None
                    ):  # Check that this is the top-level _connect call
                        # TODO: @v0.4 increase the level of this to WARNING
                        # when there's an alternative
                        with downgrade(DeprecatedException, to_level=logging.DEBUG):
                            raise DeprecatedException.from_ctx(
                                ctx,
                                f"Connected `{a}` to `{b}` by duck-typing."
                                "They should be of the same type.",
                                traceback=self.get_traceback(),
                            )

    def visitConnect_stmt(self, ctx: ap.Connect_stmtContext):
        """Connect interfaces together"""
        connectables = [self.visitConnectable(c) for c in ctx.connectable()]
        for err_cltr, (a, b) in iter_through_errors(
            itertools.pairwise(connectables),
            errors._BaseBaseUserException,
            SkipPriorFailedException,
        ):
            with err_cltr():
                self._connect(a, b, ctx)

        return NOTHING

    def visitConnectable(self, ctx: ap.ConnectableContext) -> L.ModuleInterface:
        """Return the address of the connectable object."""
        if def_stmt := ctx.pindef_stmt() or ctx.signaldef_stmt():
            (_, mif), *_ = self.visit(def_stmt)
            return mif
        elif reference_ctx := ctx.field_reference():
            ref = self.visitFieldReference(reference_ctx)
            node = self._get_referenced_node(ref, ctx)
            if not isinstance(node, L.ModuleInterface):
                raise errors.UserTypeError.from_ctx(
                    ctx,
                    f"Can't connect `{node}` because it's not a `ModuleInterface`",
                    traceback=self.get_traceback(),
                )
            return node
        else:
            raise ValueError(f"Unhandled connectable type `{ctx}`")

    def visitRetype_stmt(self, ctx: ap.Retype_stmtContext):
        from_ref = self.visitFieldReference(ctx.field_reference())
        to_ref = self.visitTypeReference(ctx.type_reference())
        from_node = self._get_referenced_node(from_ref, ctx)

        # Only Modules can be specialized (since they're the only
        # ones with specialization gifs).
        # TODO: consider duck-typing this
        if not isinstance(from_node, L.Module):
            raise errors.UserTypeError.from_ctx(
                ctx,
                f"Can't specialize `{from_node}` because it's not a `Module`",
                traceback=self.get_traceback(),
            )

        # TODO: consider extending this w/ the ability to specialize to an instance
        class_ = self._get_referenced_class(ctx, to_ref)
        with self._traceback_stack.enter(ctx):
            with self._init_node(class_) as specialized_node:
                pass

        if not isinstance(specialized_node, L.Module):
            raise errors.UserTypeError.from_ctx(
                ctx,
                f"Can't specialize with `{specialized_node}`"
                " because it's not a `Module`",
                traceback=self.get_traceback(),
            )

        # FIXME: this is an abuse of disconnect_parent. The graph isn't intended to be
        # mutable like this, and it existed only for use in traits, however the
        # alternatives I could come up with were worse:
        # - an isinstance check to additionally run `get_most_special` on Modules +
        #   more processing down the line when we want the full name of the node
        # This is only be applied when specializing to a whole class, not an instance
        try:
            parent_deets = from_node.get_parent()
            # We use from_node.get_name() rather than from_ref[-1] because we can
            # ensure this reuses the exact name after any normalization
            from_node_name = from_node.get_name()
            assert parent_deets is not None, (
                "uhh not sure how you get here without trying to replace the root node,"
                " which you shouldn't ever have access to"
            )
            parent, _ = parent_deets
            from_node.parent.disconnect_parent()
            assert isinstance(parent, L.Module)

            # We have to make sure the from_node was part of the runtime attrs
            if not any(r is from_node for r in parent.runtime.values()):
                raise errors.UserNotImplementedError.from_ctx(
                    ctx,
                    "We cannot properly specialize nodes within the base definition of"
                    " a module. This limitation mostly applies to fabll modules today.",
                    traceback=self.get_traceback(),
                )

            # Now, slot that badboi back in right where it's less-special brother's spot
            del parent.runtime[from_node_name]
            parent.add(specialized_node, name=from_node_name)

            try:
                from_node.specialize(specialized_node)
            except* L.Module.InvalidSpecializationError as ex:
                raise errors.UserException.from_ctx(
                    ctx,
                    f"Can't specialize `{from_ref}` with `{to_ref}`:\n"
                    + "\n".join(f" - {e.message}" for e in ex.exceptions),
                    traceback=self.get_traceback(),
                ) from ex
        except Exception:
            # TODO: skip further errors about this node w/ self._record_failed_node()
            raise

        return NOTHING

    def visitBlockdef(self, ctx: ap.BlockdefContext):
        """Do nothing. Handled in Surveyor."""
        return NOTHING

    def visitImport_stmt(self, ctx: ap.Import_stmtContext):
        """Do nothing. Handled in Surveyor."""
        return NOTHING

    def visitDep_import_stmt(self, ctx: ap.Dep_import_stmtContext):
        """Do nothing. Handled in Surveyor."""
        return NOTHING

    def visitAssert_stmt(self, ctx: ap.Assert_stmtContext):
        comparisons = [c for _, c in self.visitComparison(ctx.comparison())]
        for cmp in comparisons:
            if isinstance(cmp, BoolSet):
                if not cmp:
                    raise errors.UserAssertionError.from_ctx(
                        ctx,
                        "Assertion failed",
                        traceback=self.get_traceback(),
                    )
            elif isinstance(cmp, ConstrainableExpression):
                cmp.constrain()
            else:
                raise ValueError(f"Unhandled comparison type {type(cmp)}")
        return NOTHING

    # Returns fab_param.ConstrainableExpression or BoolSet
    def visitComparison(
        self, ctx: ap.ComparisonContext
    ) -> KeyOptMap[ConstrainableExpression | BoolSet]:
        exprs = [
            self.visitArithmetic_expression(c)
            for c in [ctx.arithmetic_expression()]
            + [cop.getChild(0).arithmetic_expression() for cop in ctx.compare_op_pair()]
        ]
        op_strs = [
            cop.getChild(0).getChild(0).getText() for cop in ctx.compare_op_pair()
        ]

        predicates = []
        for (lh, rh), op_str in zip(itertools.pairwise(exprs), op_strs):
            match op_str:
                # @v0.4 upgrade to error
                case "<":
                    with downgrade(
                        errors.UserNotImplementedError, to_level=logging.WARNING
                    ):
                        raise errors.UserNotImplementedError(
                            "`<` is not supported. Use `<=` instead."
                        )
                    op = LessOrEqual
                case ">":
                    with downgrade(
                        errors.UserNotImplementedError, to_level=logging.WARNING
                    ):
                        raise errors.UserNotImplementedError(
                            "`>` is not supported. Use `>=` instead."
                        )
                    op = GreaterOrEqual
                case "<=":
                    op = LessOrEqual
                case ">=":
                    op = GreaterOrEqual
                case "within":
                    op = IsSubset
                case "is":
                    op = Is
                case _:
                    # We shouldn't be able to get here with parseable input
                    raise ValueError(f"Unhandled operator `{op_str}`")

            # TODO: should we be reducing here to a series of ANDs?
            predicates.append(op(lh, rh))

        return KeyOptMap([KeyOptItem.from_kv(None, p) for p in predicates])

    def visitArithmetic_expression(
        self, ctx: ap.Arithmetic_expressionContext
    ) -> Numeric:
        if ctx.OR_OP() or ctx.AND_OP():
            raise errors.UserTypeError.from_ctx(
                ctx,
                "Logical operations are not supported",
                traceback=self.get_traceback(),
            )
            lh = self.visitArithmetic_expression(ctx.arithmetic_expression())
            rh = self.visitSum(ctx.sum_())

            if ctx.OR_OP():
                return operator.or_(lh, rh)
            else:
                return operator.and_(lh, rh)

        return self.visitSum(ctx.sum_())

    def visitSum(self, ctx: ap.SumContext) -> Numeric:
        if ctx.ADD() or ctx.MINUS():
            lh = self.visitSum(ctx.sum_())
            rh = self.visitTerm(ctx.term())

            if ctx.ADD():
                return operator.add(lh, rh)
            else:
                return operator.sub(lh, rh)

        return self.visitTerm(ctx.term())

    def visitTerm(self, ctx: ap.TermContext) -> Numeric:
        if ctx.STAR() or ctx.DIV():
            lh = self.visitTerm(ctx.term())
            rh = self.visitPower(ctx.power())

            if ctx.STAR():
                return operator.mul(lh, rh)
            else:
                return operator.truediv(lh, rh)

        return self.visitPower(ctx.power())

    def visitPower(self, ctx: ap.PowerContext) -> Numeric:
        if ctx.POWER():
            base, exp = map(self.visitFunctional, ctx.functional())
            return operator.pow(base, exp)
        else:
            return self.visitFunctional(ctx.functional(0))

    def visitFunctional(self, ctx: ap.FunctionalContext) -> Numeric:
        if ctx.name():
            name = self.visitName(ctx.name())
            operands = [self.visitBound(b) for b in ctx.bound()]
            if name == "min":
                return Min(*operands)
            elif name == "max":
                return Max(*operands)
            else:
                raise errors.UserNotImplementedError.from_ctx(
                    ctx, f"Unknown function `{name}`"
                )
        else:
            return self.visitBound(ctx.bound(0))

    def visitBound(self, ctx: ap.BoundContext) -> Numeric:
        return self.visitAtom(ctx.atom())

    def visitAtom(self, ctx: ap.AtomContext) -> Numeric:
        if ctx.field_reference():
            ref = self.visitFieldReference(ctx.field_reference())
            target = self._get_referenced_node(ref.stem, ctx)
            return self._get_param(target, ref.last, ctx)

        elif ctx.literal_physical():
            return self.visitLiteral_physical(ctx.literal_physical())

        elif group_ctx := ctx.arithmetic_group():
            assert isinstance(group_ctx, ap.Arithmetic_groupContext)
            return self.visitArithmetic_expression(group_ctx.arithmetic_expression())

        raise ValueError(f"Unhandled atom type `{ctx}`")

    def _get_unit_from_ctx(self, ctx: ParserRuleContext) -> UnitType:
        """Return a pint unit from a context."""
        unit_str = ctx.getText()
        try:
            return P.Unit(unit_str)
        except UndefinedUnitError as ex:
            raise errors.UserUnknownUnitError.from_ctx(
                ctx,
                f"Unknown unit `{unit_str}`",
                traceback=self.get_traceback(),
            ) from ex

    def visitLiteral_physical(
        self, ctx: ap.Literal_physicalContext
    ) -> Quantity_Interval:
        """Yield a physical value from a physical context."""
        if ctx.quantity():
            qty = self.visitQuantity(ctx.quantity())
            value = Single(qty)
        elif ctx.bilateral_quantity():
            value = self.visitBilateral_quantity(ctx.bilateral_quantity())
        elif ctx.bound_quantity():
            value = self.visitBound_quantity(ctx.bound_quantity())
        else:
            # this should be protected because it shouldn't be parseable
            raise ValueError
        return value

    def visitQuantity(self, ctx: ap.QuantityContext) -> Quantity:
        """Yield a physical value from an implicit quantity context."""
        raw: str = ctx.NUMBER().getText()
        if raw.startswith("0x"):
            value = int(raw, 16)
        else:
            value = float(raw)

        # Ignore the positive unary operator
        if ctx.MINUS():
            value = -value

        if unit_ctx := ctx.name():
            unit = self._get_unit_from_ctx(unit_ctx)
        else:
            unit = dimensionless

        return Quantity(value, unit)  # type: ignore

    def visitBilateral_quantity(
        self, ctx: ap.Bilateral_quantityContext
    ) -> Quantity_Interval:
        """Yield a physical value from a bilateral quantity context."""
        nominal_qty = self.visitQuantity(ctx.quantity())

        tol_ctx: ap.Bilateral_toleranceContext = ctx.bilateral_tolerance()
        tol_num = float(tol_ctx.NUMBER().getText())

        # Handle proportional tolerances
        if tol_ctx.PERCENT():
            tol_divider = 100
        elif tol_ctx.name() and tol_ctx.name().getText() == "ppm":
            tol_divider = 1e6
        else:
            tol_divider = None

        if tol_divider:
            if nominal_qty == 0:
                raise errors.UserException.from_ctx(
                    tol_ctx,
                    "Can't calculate tolerance percentage of a nominal value of zero",
                    traceback=self.get_traceback(),
                )

            # Calculate tolerance value from percentage/ppm
            tol_value = tol_num / tol_divider
            return Range.from_center_rel(nominal_qty, tol_value)

        # Ensure the tolerance has a unit
        if tol_name := tol_ctx.name():
            # In this case there's a named unit on the tolerance itself
            tol_qty = tol_num * self._get_unit_from_ctx(tol_name)
        elif nominal_qty.unitless:
            tol_qty = tol_num * dimensionless
        else:
            tol_qty = tol_num * nominal_qty.units

        # Ensure units on the nominal quantity
        if nominal_qty.unitless:
            nominal_qty = nominal_qty * HasUnit.get_units(tol_qty)

        # If the nominal has a unit, then we rely on the ranged value's unit compatibility # noqa: E501  # pre-existing
        if not nominal_qty.is_compatible_with(tol_qty):
            raise errors.UserTypeError.from_ctx(
                tol_name,
                f"Tolerance unit ({HasUnit.get_units(tol_qty)}) is not dimensionally"
                f" compatible with nominal unit ({nominal_qty.units})",
                traceback=self.get_traceback(),
            )

        return Range.from_center(nominal_qty, tol_qty)

    def visitBound_quantity(self, ctx: ap.Bound_quantityContext) -> Quantity_Interval:
        """Yield a physical value from a bound quantity context."""

        start, end = map(self.visitQuantity, ctx.quantity())

        # If only one of them has a unit, take the unit from the one which does
        if start.unitless and not end.unitless:
            start = start * end.units
        elif not start.unitless and end.unitless:
            end = end * start.units

        elif not start.is_compatible_with(end):
            # If they've both got units, let the RangedValue handle
            # the dimensional compatibility
            raise errors.UserTypeError.from_ctx(
                ctx,
                f"Tolerance unit ({end.units}) is not dimensionally"
                f" compatible with nominal unit ({start.units})",
                traceback=self.get_traceback(),
            )

        return Range(start, end)

    def visitCum_assign_stmt(self, ctx: ap.Cum_assign_stmtContext | Any):
        """
        Cumulative assignments can only be made on top of
        nothing (implicitly declared) or declared, but undefined values.

        Unlike assignments, they may not implicitly declare an attribute.
        """
        ref_dec = ctx.field_reference_or_declaration()
        assignee_ref = self.visitFieldReference(ref_dec.field_reference())
        target = self._get_referenced_node(assignee_ref, ctx)
        self.visitDeclaration_stmt(ref_dec.declaration_stmt())

        assignee = self._get_param(target, assignee_ref.last, ctx)
        value = self.visitCum_assignable(ctx.cum_assignable())

        # HACK: we have no way to check by what operator
        # the param is dynamically resolved
        # For now we assume any dynamic trait is sufficient
        if ctx.cum_operator().ADD_ASSIGN():
            assignee.alias_is(value)
        elif ctx.cum_operator().SUB_ASSIGN():
            assignee.alias_is(-value)
        else:
            # Syntax should protect from this
            raise ValueError(f"Unhandled set assignment operator {ctx}")

        # TODO: @v0.4 increase the level of this to WARNING
        # when there's an alternative
        with downgrade(DeprecatedException, to_level=logging.DEBUG):
            raise DeprecatedException(f"{ctx.cum_operator().getText()} is deprecated.")
        return NOTHING

    def visitSet_assign_stmt(self, ctx: ap.Set_assign_stmtContext):
        """
        Set cumulative assignments can only be made on top of
        nothing (implicitly declared) or declared, but undefined values.

        Unlike assignments, they may not implicitly declare an attribute.
        """
        ref_dec = ctx.field_reference_or_declaration()
        assignee_ref = self.visitFieldReference(ref_dec.field_reference())
        target = self._get_referenced_node(assignee_ref, ctx)
        self.visitDeclaration_stmt(ref_dec.declaration_stmt())

        assignee = self._get_param(target, assignee_ref.last, ctx)
        value = self.visitCum_assignable(ctx.cum_assignable())

        if ctx.OR_ASSIGN():
            assignee.constrain_superset(value)
        elif ctx.AND_ASSIGN():
            assignee.constrain_subset(value)
        else:
            # Syntax should protect from this
            raise ValueError(f"Unhandled set assignment operator {ctx}")

        # TODO: @v0.4 remove this deprecated import form
        with downgrade(DeprecatedException):
            lhs = ref_dec.field_reference().getText()
            rhs = ctx.cum_assignable().getText()
            if ctx.OR_ASSIGN():
                subset = lhs
                superset = rhs
            else:
                subset = rhs
                superset = lhs
            raise DeprecatedException(
                f"Set assignment of `{assignee}` is deprecated."
                f' Use "assert `{subset}` within `{superset}` "instead.'
            )
        return NOTHING

    def _try_get_unit_from_type_info(
        self, ctx: ap.Type_infoContext | None
    ) -> UnitType | None:
        if ctx is None:
            return None
        unit_ctx: ap.UnitContext = ctx.unit()
        if unit_ctx is None:
            return None
        # TODO: @v0.4.0: remove this shim
        unit = unit_ctx.getText()
        if unit in _declaration_domain_to_unit:
            unit = _declaration_domain_to_unit[unit]
            # TODO: consider deprecating this
        else:
            unit = self._get_unit_from_ctx(unit_ctx)

        return unit

    def _handleParameterDeclaration(
        self, ref: TypeRef, unit: UnitType, ctx: ParserRuleContext
    ):
        assert unit is not None, "Type info should be enforced by the parser"
        name = FieldRef.from_type_ref(ref).last
        param = self._ensure_param(self._current_node, name, unit, ctx)
        if param in self._param_assignments:
            declaration_after_definition = any(
                assignment.is_definition
                for assignment in self._param_assignments[param]
            )
            # TODO: @v0.4 remove this deprecated import form
            with downgrade(errors.UserKeyError):
                if declaration_after_definition:
                    raise errors.UserKeyError.from_ctx(
                        ctx,
                        f"Ignoring declaration of `{name}` "
                        "because it's already defined",
                        traceback=self.get_traceback(),
                    )
                else:
                    raise errors.UserKeyError.from_ctx(
                        ctx,
                        f"Ignoring redeclaration of `{name}`",
                        traceback=self.get_traceback(),
                    )
        else:
            self._param_assignments[param].append(
                _ParameterDefinition(
                    ref=FieldRef.from_type_ref(ref),
                    ctx=ctx,
                    traceback=self.get_traceback(),
                )
            )

    def visitDeclaration_stmt(self, ctx: ap.Declaration_stmtContext | None):
        """Handle declaration statements."""
        if ctx is None:
            return NOTHING
        ref = self.visitFieldReference(ctx.field_reference())
        if len(ref) > 1:
            raise errors.UserSyntaxError.from_ctx(
                ctx,
                f"Can't declare fields in a nested object `{ref}`",
                traceback=self.get_traceback(),
            )
        type_ref = ref.to_type_ref()
        if type_ref is None:
            raise errors.UserSyntaxError.from_ctx(
                ctx,
                f"Can't declare keyed attributes `{ref}`",
                traceback=self.get_traceback(),
            )

        # check declaration type
        unit = self._try_get_unit_from_type_info(ctx.type_info())
        if unit is not None:
            self._handleParameterDeclaration(type_ref, unit, ctx)
            return NOTHING

        assert False, "Only parameter declarations supported"

    def visitPass_stmt(self, ctx: ap.Pass_stmtContext):
        return NOTHING


bob = Bob()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.exporters.pcb.layout.heuristic_decoupling import (
    LayoutHeuristicElectricalClosenessDecouplingCaps,
)
from faebryk.exporters.pcb.layout.heuristic_pulls import (
    LayoutHeuristicElectricalClosenessPullResistors,
)
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class RP2040_ReferenceDesign(Module):
    """Minimal required design for the Raspberry Pi RP2040 microcontroller.
    Based on the official Raspberry Pi RP2040 hardware design guidlines.
    """

    class Jumper(Module):
        logic_out: F.ElectricLogic

        resistor: F.Resistor
        switch = L.f_field(F.Switch(F.Electrical))()

        def __preinit__(self):
            self.resistor.resistance.constrain_subset(
                L.Range.from_center_rel(1 * P.kohm, 0.05)
            )
            self.logic_out.set_weak(True, self).resistance.alias_is(
                self.resistor.resistance
            )
            self.logic_out.line.connect_via(
                [self.resistor, self.switch], self.logic_out.reference.lv
            )

            # TODO this is not nice
            fp = F.KicadFootprint(pin_names=["1", "2"])
            fp.add(
                F.KicadFootprint.has_kicad_identifier(
                    "Connector_PinHeader_2.54mm:PinHeader_1x02_P2.54mm_Vertical"
                )
            )
            self.switch.attach_to_footprint.attach(fp)
            # TODO make other trait for manual footprint, no pick
            self.switch.add(F.has_part_removed())

        @L.rt_field
        def single_reference(self):
            return F.has_single_electric_reference_defined(self.logic_out.reference)

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------

    # function
    rp2040: F.RP2040
    ldo: F.LDO
    clock_source: F.Crystal_Oscillator
    flash: F.Winbond_Elec_W25Q128JVSIQ

    # IO
    usb: F.USB2_0

    # UI
    boot_selector: Jumper

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://datasheets.raspberrypi.com/rp2040/hardware-design-with-rp2040.pdf"
    )

    def __preinit__(self):
        # ----------------------------------------
        #                aliasess
        # ----------------------------------------
        power_3v3 = self.add(F.ElectricPower())
        power_vbus = self.usb.usb_if.buspower

        # ----------------------------------------
        #            parametrization
        # ----------------------------------------
        # LDO
        self.ldo.output_current.constrain_ge(600 * P.mA)
        self.ldo.power_in.decoupled.decouple(owner=self).explicit(10 * P.uF)
        self.ldo.power_out.decoupled.decouple(owner=self).explicit(10 * P.uF)

        self.ldo.add(F.has_explicit_part.by_supplier("C5250992"))
        self.ldo.output_current.alias_is(1 * P.A)

        # XTAL
        xtal = self.clock_source.crystal
        xtal.load_capacitance.constrain_subset(L.Range.from_center_rel(10 * P.pF, 0.2))
        xtal.frequency.constrain_subset(L.Range.from_center_rel(12 * P.MHz, 0.05))

        self.clock_source.current_limiting_resistor.resistance.constrain_subset(
            L.Range.from_center_rel(1 * P.kohm, 0.05)
        )

        xtal.add(F.has_descriptive_properties_defined({"LCSC": "C20625731"}))
        xtal.frequency.alias_is(12 * P.MHz)
        xtal.load_capacitance.alias_is(10 * P.pF)

        # USB
        terminated_usb_data = self.add(
            self.usb.usb_if.d.terminated(), "_terminated_usb_data"
        )
        terminated_usb_data.impedance.constrain_subset(
            L.Range.from_center_rel(27.4 * P.ohm, 0.05)
        )

        # Flash
        self.flash.memory_size.constrain_subset(16 * P.Mbit)
        self.flash.decoupled.decouple(owner=self).explicit(100 * P.nF)

        # Power rails
        self.rp2040.power_io.decoupled.decouple(owner=self, count=6).explicit(
            100 * P.nF
        )
        self.rp2040.power_adc.decoupled.decouple(owner=self).explicit(100 * P.nF)
        self.rp2040.power_core.decoupled.decouple(owner=self, count=2).explicit(
            100 * P.nF
        )
        self.rp2040.power_usb_phy.decoupled.decouple(owner=self).explicit(100 * P.nF)
        self.rp2040.core_regulator.power_in.decoupled.decouple(owner=self).explicit(
            1 * P.uF
        )
        power_3v3.decoupled.decouple(owner=self).explicit(10 * P.uF)
        self.rp2040.core_regulator.power_out.decoupled.decouple(owner=self).explicit(
            1 * P.uF
        )
        # ----------------------------------------
        #              connections
        # ----------------------------------------
        # power
        power_vbus.connect_via(self.ldo, power_3v3)
        power_3v3.connect(
            self.rp2040.power_io,
            self.rp2040.power_adc,
            self.rp2040.power_usb_phy,
            self.flash.power,
        )
        power_3v3.connect_via(self.rp2040.core_regulator, self.rp2040.power_core)

        #
        self.flash.qspi.connect(self.rp2040.qspi)
        self.flash.qspi.chip_select.connect(self.boot_selector.logic_out)

        terminated_usb_data.connect(self.rp2040.usb)

        self.rp2040.xtal_if.connect(self.clock_source.xtal_if)

        LayoutHeuristicElectricalClosenessDecouplingCaps.add_to_all_suitable_modules(  # noqa: E501
            self
        )
        caps_100nF = self.get_children(
            direct_only=False,
            types=F.Capacitor,
            f_filter=lambda c: bool(
                c.capacitance.try_get_literal()
                == L.Range.from_center_rel(100 * P.nF, 0.2)
            ),
        )
        for c in caps_100nF:
            c.add(F.has_package(F.has_package.Package.C0201))

        LayoutHeuristicElectricalClosenessPullResistors.add_to_all_suitable_modules(
            self
        )

    @L.rt_field
    def pcb_layout(self):
        from faebryk.exporters.pcb.layout.absolute import LayoutAbsolute
        from faebryk.exporters.pcb.layout.extrude import LayoutExtrude
        from faebryk.exporters.pcb.layout.typehierarchy import LayoutTypeHierarchy

        Point = F.has_pcb_position.Point
        L = F.has_pcb_position.layer_type

        return F.has_pcb_layout_defined(
            layout=LayoutTypeHierarchy(
                layouts=[
                    LayoutTypeHierarchy.Level(
                        mod_type=F.RP2040,
                        layout=LayoutAbsolute(
                            Point((0, 0, 0, L.NONE)),
                        ),
                    ),
                    LayoutTypeHierarchy.Level(
                        mod_type=F.LDO,
                        layout=LayoutAbsolute(Point((10, 14, 180, L.NONE))),
                    ),
                    LayoutTypeHierarchy.Level(
                        mod_type=type(self.boot_selector),
                        layout=LayoutAbsolute(
                            Point((-1.75, -11.5, 0, L.NONE)),
                        ),
                        children_layout=LayoutExtrude(
                            vector=(3.5, 0, 90),
                        ),
                    ),
                    LayoutTypeHierarchy.Level(
                        mod_type=type(self.flash),
                        layout=LayoutAbsolute(
                            Point((-1.95, -20, 0, L.NONE)),
                        ),
                    ),
                    LayoutTypeHierarchy.Level(
                        mod_type=F.Crystal_Oscillator,
                        layout=LayoutAbsolute(
                            Point((-10, 15, 0, L.NONE)),
                        ),
                    ),
                    LayoutTypeHierarchy.Level(
                        mod_type=F.Resistor,
                        layout=LayoutExtrude(
                            base=Point((0.75, -6, 0, L.NONE)),
                            vector=(1.25, 0, 270),
                        ),
                    ),
                ]
            )
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.node import Node
from faebryk.core.parameter import Add
from faebryk.libs.library import L
from faebryk.libs.units import P


class ElectricPower(F.Power):
    class can_be_decoupled_power(F.can_be_decoupled.impl()):
        def decouple(
            self,
            owner: Module,
            count: int = 1,
        ):
            obj = self.get_obj(ElectricPower)

            capacitor = F.MultiCapacitor(count)

            # FIXME seems to cause contradictions
            capacitor.max_voltage.constrain_ge(obj.voltage * 1.5)

            obj.hv.connect_via(capacitor, obj.lv)

            name = f"decoupling_{obj.get_name(accept_no_parent=True)}"
            new_capacitor = capacitor
            # Merge
            if obj.has_trait(F.is_decoupled):
                old_capacitor = obj.get_trait(F.is_decoupled).capacitor
                capacitor = F.MultiCapacitor.from_capacitors(
                    old_capacitor,
                    new_capacitor,
                )
                name = old_capacitor.get_name(accept_no_parent=True) + "'"

            # TODO improve
            if name in owner.runtime:
                name += "_"
            owner.add(capacitor, name=name)
            obj.add(F.is_decoupled(capacitor))

            return new_capacitor

    class can_be_surge_protected_power(F.can_be_surge_protected.impl()):
        def protect(self, owner: Module):
            obj = self.get_obj(ElectricPower)
            surge_protection = F.SurgeProtection.from_interfaces(obj.lv, obj.hv)
            owner.add(
                surge_protection,
                name=f"surge_protection_{obj.get_name(accept_no_parent=True)}",
            )
            obj.add(F.is_surge_protected_defined(surge_protection))
            return surge_protection

    hv: F.Electrical
    lv: F.Electrical

    voltage = L.p_field(
        units=P.V,
        likely_constrained=True,
        domain=L.Domains.Numbers.REAL(),
        soft_set=L.Range(0 * P.V, 1000 * P.V),
        tolerance_guess=5 * P.percent,
    )
    max_current = L.p_field(
        units=P.A,
        domain=L.Domains.Numbers.REAL(),
    )
    """
    WARNING!!!
    Only for this particular power interface
    Does not propagate to connections
    """
    bus_max_current_consumption_sum = L.p_field(
        units=P.A, domain=L.Domains.Numbers.REAL()
    )
    """
    Summed current for all connected power interfaces
    Only available after resolve_bus_parameters
    """

    surge_protected: can_be_surge_protected_power
    decoupled: can_be_decoupled_power

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(self)

    def fused(self, attach_to: Node | None = None):
        fused_power = type(self)()
        fuse = fused_power.add(F.Fuse())

        fused_power.hv.connect_via(fuse, self.hv)
        fused_power.lv.connect(self.lv)

        self.connect_shallow(fused_power)

        fuse.trip_current.constrain_subset(
            self.max_current * L.Range.from_center_rel(1.0, 0.1)
        )
        # TODO maybe better bus_consumption
        fused_power.max_current.constrain_le(fuse.trip_current)

        if attach_to is not None:
            attach_to.add(fused_power)

        return fused_power

    def __preinit__(self) -> None:
        ...
        # self.voltage.alias_is(
        #    self.hv.potential - self.lv.potential
        # )
        self.voltage.add(F.is_bus_parameter())
        self.bus_max_current_consumption_sum.add(
            F.is_bus_parameter(reduce=(self.max_current, Add))
        )

        self.lv.add(F.has_net_name("gnd"))

    @property
    def vcc(self) -> F.Electrical:
        """Higher-voltage side of the power interface."""
        return self.hv

    @property
    def gnd(self) -> F.Electrical:
        """Lower-voltage side of the power interface."""
        return self.lv



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class CBM9002A_56ILG_ReferenceDesign(Module):
    """
    Minimal working example for the CBM9002A_56ILG
    """

    class ResetCircuit(Module):
        """low-pass and protection for reset"""

        diode: F.Diode
        cap: F.Capacitor
        logic: F.ElectricLogic

        def __preinit__(self):
            self.logic.line.connect_via(self.diode, self.logic.reference.hv)
            self.logic.pulled.pull(up=True, owner=self)
            self.logic.line.connect_via(self.cap, self.logic.reference.lv)

            self.cap.capacitance.constrain_subset(
                L.Range.from_center_rel(1 * P.uF, 0.05)
            )

            self.diode.forward_voltage.constrain_subset(L.Range(715 * P.mV, 1.5 * P.V))
            self.diode.reverse_leakage_current.constrain_le(1 * P.uA)
            self.diode.current.constrain_subset(
                L.Range.from_center_rel(300 * P.mA, 0.05)
            )
            self.diode.current.constrain_ge(1 * P.A)

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    mcu: F.CBM9002A_56ILG
    oscillator: F.Crystal_Oscillator
    reset_circuit: ResetCircuit

    PA = L.list_field(8, F.ElectricLogic)
    PB = L.list_field(8, F.ElectricLogic)
    PD = L.list_field(8, F.ElectricLogic)
    usb: F.USB2_0
    i2c: F.I2C

    avcc: F.ElectricPower
    vcc: F.ElectricPower

    rdy = L.list_field(2, F.ElectricLogic)
    ctl = L.list_field(3, F.ElectricLogic)
    reset: F.ElectricLogic
    wakeup: F.ElectricLogic

    ifclk: F.ElectricLogic
    clkout: F.ElectricLogic
    xtalin: F.Electrical
    xtalout: F.Electrical

    # ----------------------------------------
    #                traits
    # ----------------------------------------

    # ----------------------------------------
    #                connections
    # ----------------------------------------
    def __preinit__(self):
        self.connect_interfaces_by_name(self.mcu, allow_partial=True)

        # crystal oscillator
        self.oscillator.xtal_if.gnd.connect(self.vcc.lv)
        self.oscillator.xtal_if.xin.connect(self.xtalin)
        self.oscillator.xtal_if.xout.connect(self.xtalout)

        self.reset_circuit.logic.connect(self.mcu.reset)

        self.avcc.decoupled.decouple(owner=self)
        self.vcc.decoupled.decouple(owner=self)

        # ----------------------------------------
        #               Parameters
        # ----------------------------------------

        self.oscillator.crystal.frequency.constrain_subset(
            L.Range.from_center_rel(24 * P.Mhertz, 0.05)
        )
        self.oscillator.crystal.frequency_tolerance.constrain_subset(
            L.Range(0 * P.ppm, 20 * P.ppm)
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from pathlib import Path
from typing import cast

import atopile.config as config
from atopile import errors
from faebryk.libs.backend.packages.api import PackagesAPIClient
from faebryk.libs.package.dist import Dist, DistValidationError
from faebryk.libs.util import (
    DAG,
    clone_repo,
    duplicates,
    find_or,
    md_list,
    not_none,
    robustly_rm_dir,
)

logger = logging.getLogger(__name__)


class ProjectDependency:
    def __init__(
        self,
        spec: config.DependencySpec,
        pcfg: config.ProjectConfig | None = None,
        gcfg: config.Config | None = None,
    ):
        self.spec = spec
        self.dist: Dist | None = None
        self.gcfg = gcfg or config.config
        self.pcfg = pcfg or self.gcfg.project

        self.cfg: config.ProjectConfig | None = None

    # TODO see __eq__
    def __hash__(self) -> int:
        # used for BFS resolution
        return hash(self.spec.identifier)

    # TODO don't make property of object, but use as comparator where needed
    def __eq__(self, other: object) -> bool:
        # used for BFS resolution
        if not isinstance(other, ProjectDependency):
            return False
        if self.spec.identifier == other.spec.identifier:
            # TODO do some asserts about the equivalence
            return True
        return False

    @property
    def identifier(self) -> str:
        if self.spec.identifier is not None:
            return self.spec.identifier
        assert self.dist is not None
        return self.dist.identifier

    @property
    def project_config(self) -> config.ProjectConfig:
        if self.cfg is not None:
            return self.cfg
        assert self.dist is not None
        return self.dist.manifest

    @property
    def target_path(self) -> Path:
        # TODO don't really like using identifier as import path
        # would be nicer to use source name and indirect imports
        return self.gcfg.project.paths.modules / self.identifier

    @property
    def cache_dir(self) -> Path:
        return self.gcfg.project.paths.modules / ".cache"

    def add_to_manifest(self):
        config.ProjectConfig.set_or_add_dependency(self.gcfg, self.spec)

    def remove_from_manifest(self):
        config.ProjectConfig.remove_dependency(self.gcfg, self.spec)

    def load_dist(self):
        if self.dist is not None:
            return
        # TODO implement cache
        temp_dir = self.cache_dir
        if isinstance(self.spec, (config.FileDependencySpec, config.GitDependencySpec)):
            if isinstance(self.spec, config.FileDependencySpec):
                path = self.spec.path
                if not path.exists():
                    raise FileNotFoundError(
                        f"Local dependency path {path} does not exist for"
                        f" {self.spec.identifier}"
                    )
            else:
                repo_cache = Path(temp_dir) / ".git_repo.cache"
                if repo_cache.exists():
                    robustly_rm_dir(repo_cache)
                path = clone_repo(
                    self.spec.repo_url,
                    clone_target=repo_cache,
                    ref=self.spec.ref,
                )
                if self.spec.path_within_repo:
                    path = path / self.spec.path_within_repo

            dist = Dist.build_dist(
                cfg=path,
                output_path=Path(temp_dir),
            )
            self.spec.identifier = dist.identifier

        elif isinstance(self.spec, config.RegistryDependencySpec):
            api = PackagesAPIClient()
            dist = api.release_dist(
                self.spec.identifier,
                Path(temp_dir),
                version=self.spec.release,
            )
            self.spec.release = dist.version
        else:
            raise NotImplementedError(f"Loading dist for {self.spec} not implemented")
        self.dist = dist

    @property
    def direct_dependencies(self) -> list["ProjectDependency"]:
        return [
            not_none(ProjectDependency(spec, pcfg=self.project_config, gcfg=self.gcfg))
            for spec in self.project_config.dependencies or []
        ]

    def try_load(self):
        if self.cfg:
            return
        if not self.target_path:
            return
        if not self.target_path.exists():
            return
        self.cfg = config.ProjectConfig.from_path(self.target_path)

    def __str__(self) -> str:
        return f"{type(self).__name__}(spec={self.spec},path={self.target_path})"

    def __repr__(self) -> str:
        return str(self)


class ProjectDependencies:
    def __init__(
        self,
        pcfg: config.ProjectConfig | None = None,
        install_missing: bool = False,
        clean_unmanaged_dirs: bool = False,
    ):
        if pcfg is None:
            pcfg = config.config.project
        self.pcfg = pcfg

        self.direct_deps = {
            ProjectDependency(spec, pcfg=pcfg) for spec in pcfg.dependencies or []
        }
        self.dag = self.resolve_dependencies()

        if install_missing:
            self.install_missing_dependencies()
        if clean_unmanaged_dirs:
            self.clean_unmanaged_directories()

    @property
    def all_deps(self) -> set[ProjectDependency]:
        return self.dag.values

    def reload(self):
        self.__init__(pcfg=self.pcfg)

    @property
    def not_installed_dependencies(self) -> set[ProjectDependency]:
        return {dep for dep in self.all_deps if not dep.cfg}

    def install_missing_dependencies(self):
        for dep in self.not_installed_dependencies:
            assert dep.dist is not None
            dep.dist.install(dep.target_path)

    def clean_unmanaged_directories(self):
        module_dir = self.pcfg.paths.modules
        if not module_dir.exists():
            return
        local_dep_dirs = set()
        for owner_dir in module_dir.iterdir():
            if owner_dir.name.startswith("."):
                continue
            if not owner_dir.is_dir():
                continue
            for dep_dir in owner_dir.iterdir():
                if not dep_dir.is_dir():
                    continue
                local_dep_dirs.add(dep_dir.relative_to(module_dir))

        dep_dirs = {
            not_none(dep.target_path).relative_to(module_dir) for dep in self.all_deps
        }
        # Get the first two parts of the path (owner/repo) to use as prefixes
        dep_dir_prefixes = {Path(*dep_dir.parts[:2]) for dep_dir in dep_dirs}

        unmanaged_dirs = local_dep_dirs - dep_dir_prefixes
        if not unmanaged_dirs:
            return
        logger.warning(
            f"Removing unmanaged module directories: {unmanaged_dirs}",
        )
        for unmanaged_dir in unmanaged_dirs:
            robustly_rm_dir(module_dir / unmanaged_dir)

    def resolve_dependencies(self):
        dag = DAG[ProjectDependency]()
        # TODO: can be replaced with dag.values
        all_deps: set[ProjectDependency] = set()

        # Good old BFS for dependency resolution
        deps_to_process: list[tuple[ProjectDependency | None, ProjectDependency]] = [
            (None, dep) for dep in self.direct_deps
        ]
        while deps_to_process:
            to_add = []
            for parent, dep in deps_to_process:
                dups = all_deps.intersection({dep})
                assert len(dups) <= 1
                dup = dups.pop() if dups else None
                if dup:
                    if dup.spec != dep.spec:
                        # TODO better error
                        raise errors.UserException(
                            f"Incompatible dependency specs in tree: {dep.spec}"
                        )
                    assert parent is not None, "Can't have duplicates in root"
                    dag.add_edge(parent, dup)
                    continue

                dep.try_load()
                if dep.cfg is None:
                    dep.load_dist()
                to_add.extend((dep, child) for child in dep.direct_dependencies)
                all_deps.add(dep)
                if parent is not None:
                    dag.add_edge(parent, dep)
                else:
                    dag.add_or_get(dep)

            deps_to_process.clear()
            deps_to_process.extend(to_add)

        if dag.contains_cycles:
            # TODO better error
            raise errors.UserException("Cycle in dependencies")

        assert dag.values == all_deps
        return dag

    def remove(self, spec: config.DependencySpec):
        raise NotImplementedError("Removing dependencies is not implemented")

    def install_from_spec_to_manifest(
        self, dep: ProjectDependency, upgrade: bool = False
    ):
        # loaded identifier from dist
        identifier = dep.identifier

        existing_dep = find_or(
            self.dag.values,
            lambda d: not_none(d).identifier == identifier,
            default=None,
        )

        if existing_dep is not None:
            if type(existing_dep.spec) is not type(dep.spec):
                raise errors.UserException(
                    f"Cannot install {identifier} as it is already installed "
                    f"from a different source: {existing_dep.spec}"
                )
            if (
                isinstance(dep.spec, config.RegistryDependencySpec)
                and isinstance(existing_dep.spec, config.RegistryDependencySpec)
                and dep.spec.release != existing_dep.spec.release
            ):
                if existing_dep not in self.direct_deps:
                    raise errors.UserException(
                        f"Cannot install {identifier} as it is already installed "
                        f"with a different version: {existing_dep.spec.release}"
                        f"from a transitive dependency"
                    )
                if not upgrade:
                    raise errors.UserException(
                        f"Cannot install {identifier} as it is already installed "
                        f"with a different version: {existing_dep.spec.release}"
                        f"Use --upgrade to install anyway"
                    )
            if isinstance(
                existing_dep.spec,
                (config.FileDependencySpec, config.GitDependencySpec),
            ):
                if not upgrade:
                    raise errors.UserException(
                        f"Cannot install {identifier} as it is already installed "
                        f"Use --upgrade to install anyway"
                    )

        target_path = dep.target_path
        if target_path.exists():
            robustly_rm_dir(target_path)

        dep.load_dist()
        assert dep.dist is not None

        logger.info(f"Installing {identifier} to {target_path}")
        dep.dist.install(target_path)
        dep.add_to_manifest()

    def add_dependencies(self, *specs: config.DependencySpec, upgrade: bool = False):
        # Load specs and fetch dists if needed
        new_deps = []
        for spec in specs:
            dep = ProjectDependency(spec)
            if dep.spec.identifier is None:
                try:
                    dep.load_dist()
                except DistValidationError as e:
                    raise errors.UserException(
                        f"Could not load distribution for {spec}: {e}"
                    ) from e
            new_deps.append(dep)

        # Check for duplicates
        dups = duplicates(new_deps, lambda x: x.spec.identifier)
        if dups:
            raise errors.UserException(f"Duplicate specs provided: {dups}")

        # Check if in manifest and/or already installed
        # install locally and in manifest
        for dep in new_deps:
            self.install_from_spec_to_manifest(dep, upgrade=upgrade)

        self.reload()

    def remove_dependencies(self, *identifiers: str):
        all_deps_by_identifier = {dep.identifier: dep for dep in self.all_deps}

        to_remove = {
            identifier: all_deps_by_identifier.get(identifier)
            for identifier in identifiers
            if identifier in all_deps_by_identifier
        }

        not_found = [item for item in to_remove.items() if item[1] is None]
        if not_found:
            raise errors.UserException(
                "Could not find dependencies to remove:"
                f" {md_list(dict(not_found).keys())}"
            )
        to_remove_deps = cast(set[ProjectDependency], set(to_remove.values()))

        non_root_deps = to_remove_deps - self.direct_deps
        if non_root_deps:
            raise errors.UserException(
                "Cannot remove dependencies"
                " that are not direct dependencies of this project:"
                f"\n{md_list(non_root_deps)}"
            )

        uninstall = set[ProjectDependency]()
        no_uninstall = set[ProjectDependency]()
        for dep in to_remove_deps:
            transitive_removals = self.dag.all_parents(dep)
            # direct project dependencies that depend on a to_be_removed dependency
            # and are not to_be_removed themselves
            non_removable = (transitive_removals - to_remove_deps) & self.direct_deps
            if non_removable:
                no_uninstall.add(dep)
                continue
            uninstall.add(dep)

        if no_uninstall:
            logger.info(
                f"Will remove dependencies but not uninstall following packages as they"
                f" are required by downstream dependencies:"
                f"\n{md_list([dep.identifier for dep in no_uninstall])}",
                extra={"markdown": True},
            )

        if uninstall:
            logger.info(
                f"Uninstalling and removing:"
                f"\n{md_list([dep.identifier for dep in uninstall])}",
                extra={"markdown": True},
            )
        for dep in to_remove_deps:
            dep.remove_from_manifest()

        for dep in uninstall:
            if dep.target_path.exists():
                robustly_rm_dir(dep.target_path)

        # reload and clean orphaned packages
        self.reload()
        self.clean_unmanaged_directories()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from abc import abstractmethod

from faebryk.core.module import Module
from faebryk.core.parameter import Parameter


class has_resistance(Module.TraitT):
    @abstractmethod
    def get_resistance(self) -> Parameter: ...



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.picker.picker import DescriptiveProperties
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class Winbond_Elec_W25Q128JVSIQ(F.SPIFlash):
    """
    TODO: Docstring describing your module

    128Mbit SOIC-8-208mil
    NOR FLASH ROHS
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    @L.rt_field
    def has_pick(self):
        return F.has_explicit_part.by_supplier("C97521", supplier_id="lcsc")

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )
    descriptive_properties = L.f_field(F.has_descriptive_properties_defined)(
        {
            DescriptiveProperties.manufacturer: "Winbond Elec",
            DescriptiveProperties.partno: "W25Q128JVSIQ",
        }
    )
    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://wmsc.lcsc.com/wmsc/upload/file/pdf/v2/lcsc/1811142111_Winbond-Elec-W25Q128JVSIQ_C97521.pdf"
    )

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.qspi.chip_select.line: ["CS#"],
                self.qspi.data[0].line: ["DO"],
                self.qspi.data[2].line: ["IO2"],
                self.power.lv: ["GND"],
                self.qspi.data[1].line: ["DI"],
                self.qspi.clock.line: ["CLK"],
                self.qspi.data[3].line: ["IO3"],
                self.power.hv: ["VCC"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )

    @L.rt_field
    def decoupled(self):
        return F.can_be_decoupled_rails(self.power)

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        pass



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import faebryk.library._F as F
from faebryk.core.module import Module


class has_pcb_position_defined_relative(F.has_pcb_position.impl()):
    def __init__(self, position_relative: F.has_pcb_position.Point, to: Module) -> None:
        super().__init__()
        self.position_relative = position_relative
        self.to = to

    def get_position(self) -> F.has_pcb_position.Point:
        from faebryk.libs.geometry.basic import Geometry

        return Geometry.abs_pos(
            self.to.get_trait(F.has_pcb_position).get_position(), self.position_relative
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class TD541S485H(Module):
    power: F.ElectricPower
    power_iso_in: F.ElectricPower
    power_iso_out: F.ElectricPower
    uart: F.UART_Base
    rs485: F.RS485HalfDuplex
    read_enable: F.ElectricLogic
    write_enable: F.ElectricLogic

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )

    @L.rt_field
    def attach_to_footprint(self):
        x = self
        return F.can_attach_to_footprint_via_pinmap(
            {
                "1": x.power.lv,
                "2": x.power.hv,
                "3": x.uart.rx.line,
                "4": x.read_enable.line,
                "5": x.write_enable.line,
                "6": x.uart.tx.line,
                "7": x.power.hv,
                "8": x.power.lv,
                "9": x.power_iso_out.lv,
                "10": x.power_iso_out.hv,
                "13": x.rs485.diff_pair.n.line,
                "14": x.rs485.diff_pair.p.line,
                "15": x.power_iso_in.hv,
                "16": x.power_iso_in.lv,
            }
        )

    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://www.mornsun-power.com/public/uploads/pdf/TD(H)541S485H.pdf"
    )

    def __preinit__(self):
        # FIXME
        # self.power.decoupled.decouple()
        # self.power_iso_in.decoupled.decouple()
        # self.power_iso_out.decoupled.decouple()

        self.power_iso_in.lv.connect(self.power_iso_out.lv)
        # TODO tolerance
        self.power_iso_out.voltage.constrain_superset(5 * P.V)

        F.ElectricLogic.connect_all_module_references(
            self,
            exclude=[
                self.power,
                self.uart,
                self.read_enable,
                self.write_enable,
            ],
        )

        # TODO: ugly
        self.rs485.diff_pair.n.reference.connect(self.power_iso_out)
        self.rs485.diff_pair.p.reference.connect(self.power_iso_out)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from typing import Any, Callable

from typing_extensions import Self

from faebryk.libs.util import ConfigFlag

logger = logging.getLogger(__name__)

ID_REPR = ConfigFlag("ID_REPR", False, "Add object id to repr")


class FaebrykLibObject:
    def __init__(self) -> None: ...

    def builder(self, op: Callable[[Self], Any]) -> Self:
        op(self)
        return self


class Namespace:
    """Marker class for namespace objects."""

    pass



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F


class can_switch_power_defined(F.can_switch_power.impl()):
    def __init__(
        self,
        in_power: F.ElectricPower,
        out_power: F.ElectricPower,
        in_logic: F.ElectricLogic,
    ) -> None:
        super().__init__()

        self.in_power = in_power
        self.out_power = out_power
        self.in_logic = in_logic

        out_power.voltage.alias_is(in_power.voltage)

    def get_logic_in(self) -> F.ElectricLogic:
        return self.in_logic

    def get_in(self) -> F.ElectricPower:
        return self.in_power

    def get_out(self) -> F.ElectricPower:
        return self.out_power



# pylint: disable=logging-fstring-interpolation

"""
`ato view`
"""

import logging
from typing import Annotated

import typer

from atopile import errors

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


def view(
    entry: Annotated[str | None, typer.Argument()] = None,
    build: Annotated[list[str], typer.Option("--build", "-b", envvar="ATO_BUILD")] = [],
    target: Annotated[
        list[str], typer.Option("--target", "-t", envvar="ATO_TARGET")
    ] = [],
    option: Annotated[
        list[str], typer.Option("--option", "-o", envvar="ATO_OPTION")
    ] = [],
):
    """
    View a block diagram or schematic of your project.
    """
    raise errors.UserNotImplementedError("View is not yet implemented.")



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Literal
from urllib.parse import urlparse

import requests
from dataclasses_json.api import dataclass_json
from github_oidc.client import get_actions_header
from pydantic.networks import HttpUrl

from atopile.config import config
from faebryk.libs.package.dist import Dist
from faebryk.libs.util import indented_container, once

logger = logging.getLogger(__name__)


class _Models:
    class Publish:
        @dataclass_json
        @dataclass(frozen=True)
        class Request:
            identifier: str
            package_version: str
            package_size: int
            manifest: dict[str, Any]
            """
            contents of the `ato.yaml` file
            """

        @dataclass_json
        @dataclass(frozen=True)
        class Response:
            @dataclass_json
            @dataclass(frozen=True)
            class Info:
                url: HttpUrl
                fields: dict[str, str]

            status: Literal["ok"]
            release_id: str
            upload_info: Info

    class PublishUploadComplete:
        @dataclass_json
        @dataclass(frozen=True)
        class Request:
            release_id: str

        @dataclass_json
        @dataclass(frozen=True)
        class Response:
            status: Literal["ok"]
            url: str

    class Package:
        @dataclass_json
        @dataclass(frozen=True)
        class Response:
            @dataclass_json
            @dataclass(frozen=True)
            class Info:
                version: str

            info: Info

    class Release:
        @dataclass_json
        @dataclass(frozen=True)
        class Response:
            @dataclass_json
            @dataclass(frozen=True)
            class Info:
                # @dataclass_json
                # @dataclass(frozen=True)
                # class Dependencies:
                #     @dataclass_json
                #     @dataclass(frozen=True)
                #     class Dependency:
                #         identifier: str
                #         version: str

                #     requires: list[Dependency]

                # dependencies: Dependencies
                download_url: str
                # requires_atopile: str
                filename: str

            info: Info


class Errors:
    class PackagesApiError(Exception): ...

    class PackagesApiHTTPError(Exception):
        def __init__(self, error: requests.exceptions.HTTPError, detail: str):
            super().__init__()
            self.error = error
            self.response = error.response
            self.detail = detail

        @classmethod
        def from_http(cls, error: "Errors.PackagesApiHTTPError"):
            return cls(error.error, error.detail)

        @property
        def code(self) -> int:
            return self.response.status_code

        def __str__(self) -> str:
            return f"{type(self).__name__}: {self.code} {self.detail}"

    class AuthenticationError(PackagesApiError): ...

    class PackageNotFoundError(PackagesApiHTTPError):
        def __init__(
            self,
            error: requests.exceptions.HTTPError,
            detail: str,
            package_identifier: str,
        ):
            super().__init__(error, detail)
            self.package_identifier = package_identifier

        def __str__(self) -> str:
            return f"{type(self).__name__}: {self.package_identifier}: {self.detail}"

        @classmethod
        def from_http(
            cls, error: "Errors.PackagesApiHTTPError", package_identifier: str
        ):
            return cls(error.error, error.detail, package_identifier)

    class InvalidPackageIdentifierError(PackagesApiHTTPError):
        def __init__(
            self,
            error: requests.exceptions.HTTPError,
            detail: str,
            package_identifier: str,
        ):
            super().__init__(error, detail)
            self.package_identifier = package_identifier

        def __str__(self) -> str:
            return f"{type(self).__name__}: {self.package_identifier}: {self.detail}"

        @classmethod
        def from_http(
            cls, error: "Errors.PackagesApiHTTPError", package_identifier: str
        ):
            return cls(error.error, error.detail, package_identifier)

    class ReleaseNotFoundError(PackagesApiHTTPError):
        def __init__(
            self,
            error: requests.exceptions.HTTPError,
            detail: str,
            package_identifier: str,
            release: str,
        ):
            super().__init__(error, detail)
            self.package_identifier = package_identifier
            self.release = release

        def __str__(self) -> str:
            return (
                f"{type(self).__name__}: {self.package_identifier}@{self.release}:"
                f" {self.detail}"
            )

        @classmethod
        def from_http(
            cls,
            error: "Errors.PackagesApiHTTPError",
            package_identifier: str,
            release: str,
        ):
            return cls(error.error, error.detail, package_identifier, release)

    class ReleaseAlreadyExistsError(PackagesApiHTTPError): ...


class PackagesAPIClient:
    @dataclass
    class ApiConfig:
        api_url: str = config.project.services.packages.url

    @property
    @once
    def _cfg(self) -> ApiConfig:
        return self.ApiConfig()

    def __init__(self):
        self._client = requests.Session()

    def _get(
        self,
        url: str,
        timeout: float = 10,
    ) -> requests.Response:
        response = self._client.get(f"{self._cfg.api_url}{url}", timeout=timeout)
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            try:
                detail = response.json()["detail"]
            except (requests.JSONDecodeError, KeyError):
                detail = response.text
            raise Errors.PackagesApiHTTPError(e, detail) from e
        return response

    def _post(
        self,
        url: str,
        data: Any,
        timeout: float = 10,
        authenticate: bool = False,
    ) -> requests.Response:
        headers = {}
        if authenticate:
            headers |= self._authenticate()

        response = self._client.post(
            f"{self._cfg.api_url}{url}",
            json=data,
            timeout=timeout,
            headers=headers,
        )
        try:
            response.raise_for_status()
            assert response.json()["status"] == "ok"
        except requests.exceptions.HTTPError as e:
            try:
                detail = response.json()["detail"]
            except (requests.JSONDecodeError, KeyError):
                detail = response.text
            raise Errors.PackagesApiHTTPError(e, detail) from e
        return response

    def _upload(
        self,
        url: str,
        file_path: Path,
        data: dict[str, str],
        timeout: float = 10,
        skip_verify: bool = False,
    ) -> requests.Response:
        response = self._client.post(
            url,
            data=data,
            files={
                "file": (file_path.name, file_path.read_bytes()),
                "Content-Type": "application/zip",
            },
            timeout=timeout,
            verify=not skip_verify,
        )
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            try:
                detail = response.json()["detail"]
            except (requests.JSONDecodeError, KeyError):
                detail = response.text
            raise Errors.PackagesApiHTTPError(e, detail) from e
        return response

    def _authenticate(self) -> dict[str, str]:
        """
        Currently, the only supported authentication method is Github Actions OIDC.
        """
        try:
            return get_actions_header(urlparse(self._cfg.api_url).netloc)
        except Exception as e:
            raise Errors.AuthenticationError(e) from e

    def publish(
        self, identifier: str, version: str, dist: Dist, skip_auth: bool = False
    ) -> _Models.PublishUploadComplete.Response:
        """
        Publish a package to the package registry.

        Only works in Github Actions.
        """
        # Get filesize of dist.path
        filesize = dist.path.stat().st_size

        ## Request upload
        try:
            r = self._post(
                "/v1/publish",
                data=_Models.Publish.Request(
                    identifier=identifier,
                    package_version=version,
                    package_size=filesize,
                    manifest=dist.manifest.model_dump(mode="json"),
                ).to_dict(),  # type: ignore
                authenticate=not skip_auth,
            )
        except Errors.PackagesApiHTTPError as e:
            if e.code == 409:
                raise Errors.ReleaseAlreadyExistsError.from_http(e) from e
            raise

        response = _Models.Publish.Response.from_dict(r.json())  # type: ignore

        ## Upload the package
        try:
            r = self._upload(
                url=response.upload_info.url,
                file_path=dist.path,
                data=response.upload_info.fields,
                skip_verify=skip_auth,
            )
        except Errors.PackagesApiHTTPError:
            raise

        ## Confirm upload
        try:
            r = self._post(
                "/v1/publish/upload-complete",
                data=_Models.PublishUploadComplete.Request(
                    release_id=response.release_id,
                ).to_dict(),  # type: ignore
                authenticate=not skip_auth,
            )
        except Errors.PackagesApiHTTPError:
            raise

        response = _Models.PublishUploadComplete.Response.from_dict(r.json())  # type: ignore
        return response

    def package(
        self, identifier: str, version: str | None = None
    ) -> _Models.Release.Response:
        """
        Get a package from the package registry.
        """
        if version is None:
            try:
                r = self._get(f"/v1/package/{identifier}")
            except Errors.PackagesApiHTTPError as e:
                if e.code == 404:
                    raise Errors.PackageNotFoundError.from_http(e, identifier) from e
                if e.code == 422:
                    raise Errors.InvalidPackageIdentifierError.from_http(
                        e, identifier
                    ) from e
                raise
            response = _Models.Package.Response.from_dict(r.json())  # type: ignore
            assert isinstance(response, _Models.Package.Response)
            version = response.info.version

        try:
            r = self._get(f"/v1/package/{identifier}/releases/{version}")
        except Errors.PackagesApiHTTPError as e:
            if e.code == 404:
                raise Errors.ReleaseNotFoundError.from_http(
                    e, identifier, version
                ) from e
            raise
        try:
            response = _Models.Release.Response.from_dict(r.json())  # type: ignore
        except Exception:
            logger.error(indented_container(r.json(), recursive=True))
            raise
        return response

    def release_dist(
        self, identifier: str, output_path: Path, version: str | None = None
    ) -> Dist:
        release = self.package(identifier, version)
        url = release.info.download_url
        filepath = output_path / release.info.filename
        filepath.parent.mkdir(parents=True, exist_ok=True)
        # use requests to download the file to output_path
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            with filepath.open("wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)

        return Dist(filepath)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from pathlib import Path

import faebryk.library._F as F
from faebryk.core.graph import Graph, GraphFunctions
from faebryk.core.module import Module
from faebryk.core.node import Node
from faebryk.libs.app.erc import simple_erc
from faebryk.libs.exceptions import UserDesignCheckException, accumulate, downgrade
from faebryk.libs.kicad.fileformats_latest import C_kicad_drc_report_file
from faebryk.libs.units import to_si_str
from faebryk.libs.util import groupby, md_list

logger = logging.getLogger(__name__)


class RequiresExternalUsageNotFulfilled(UserDesignCheckException):
    def __init__(self, nodes: list[Node]):
        self.nodes = nodes
        super().__init__(
            f"Nodes requiring external usage but not used externally: "
            f"{', '.join(mif.get_full_name() for mif in nodes)}"
        )


# TODO: convert to trait-based check
def check_requires_external_usage(app: Module, G: Graph):
    unfulfilled = []
    for node, trait in GraphFunctions(G).nodes_with_trait(F.requires_external_usage):
        if not trait.fulfilled:
            # Don't check the app module itself
            if app is node:
                continue
            unfulfilled.append(node)
    if unfulfilled:
        raise RequiresExternalUsageNotFulfilled(unfulfilled)


def check_design(G: Graph):
    # TODO: split checks by stage
    with accumulate(UserDesignCheckException) as accumulator:
        for _, trait in GraphFunctions(G).nodes_with_trait(F.implements_design_check):
            with accumulator.collect():
                try:
                    trait.check()
                except F.implements_design_check.MaybeUnfulfilledCheckException as e:
                    with downgrade(UserDesignCheckException):
                        raise UserDesignCheckException.from_nodes(
                            str(e), e.nodes
                        ) from e
                except F.implements_design_check.UnfulfilledCheckException as e:
                    raise UserDesignCheckException.from_nodes(str(e), e.nodes) from e


def run_pre_build_checks(app: Module, G: Graph):
    check_requires_external_usage(app, G)
    check_design(G)
    simple_erc(G)


def run_post_build_checks(app: Module, G: Graph):
    check_design(G)


def run_post_pcb_checks(app: Module, G: Graph, pcb: Path):
    run_drc(app, G, pcb)


type Violation = C_kicad_drc_report_file.C_Violation


class DrcException(UserDesignCheckException):
    def __init__(
        self,
        shorts: list[Violation],
        unconnected: list[Violation],
        units: str,
    ):
        self.shorts = shorts
        self.unconnected = unconnected
        self.units = units
        super().__init__(
            f"{type(self).__name__} ("
            f"{len(self.shorts)} shorts,"
            f"{len(self.unconnected)} unconnected"
            f")"
        )

    def pretty_violation(self, violation: Violation):
        def _convert_coord(c):
            x, y = (to_si_str(subcoord, self.units) for subcoord in (c.x, c.y))
            return f"({x},{y})"

        return {
            violation.description: [
                f"{i.description} @{_convert_coord(i.pos)}" for i in violation.items
            ]
        }

    def pretty(self) -> str:
        out = ""
        if self.shorts:
            out += "\n\nShorts\n"
            out += md_list(
                [self.pretty_violation(v) for v in self.shorts],
                recursive=True,
            )
        if self.unconnected:
            out += "\n\nMissing connections\n"
            out += md_list(
                [self.pretty_violation(v) for v in self.unconnected],
                recursive=True,
            )
        return out


# TODO: convert to trait-based check
def run_drc(app: Module, G: Graph, pcb: Path):
    from faebryk.libs.kicad.drc import run_drc as run_drc_kicad

    drc_report = run_drc_kicad(pcb)

    grouped = groupby(drc_report.violations, lambda v: v.type)
    not_connected = drc_report.unconnected_items

    shorts = grouped.get(C_kicad_drc_report_file.C_Violation.C_Type.shorting_items, [])
    if shorts or not_connected:
        raise DrcException(shorts, not_connected, drc_report.coordinate_units)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from dataclasses import dataclass

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.node import Node
from faebryk.exporters.pcb.layout.heuristic_decoupling import Params, place_next_to
from faebryk.exporters.pcb.layout.layout import Layout
from faebryk.libs.util import not_none

logger = logging.getLogger(__name__)


@dataclass(frozen=True, eq=True)
class LayoutNextToPin(Layout):
    interface: F.Electrical
    distance_between_pad_edges: float = 1
    extra_rotation_of_footprint: float = 0

    def apply(self, *node: Node):
        """
        Tip: Make sure at least one parent of node has an absolute position defined
        """
        # Remove nodes that have a position defined
        node = tuple(n for n in node if not n.has_trait(F.has_pcb_position))

        for n in node:
            assert isinstance(n, Module)
            electrical = not_none(self.interface)
            logger.debug(f"Placing {n} next to {electrical}")
            place_next_to(
                electrical,
                n,
                route=True,
                params=Params(
                    self.distance_between_pad_edges, self.extra_rotation_of_footprint
                ),
            )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import csv
import logging
import os
import re
from dataclasses import dataclass
from pathlib import Path

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.picker.picker import DescriptiveProperties

logger = logging.getLogger(__name__)


@dataclass
class BOMLine:
    Designator: str
    Footprint: str
    Quantity: int
    Value: str
    Manufacturer: str
    Partnumber: str
    LCSC_Partnumber: str

    def to_dict(self) -> dict[str, str]:
        out = vars(self)
        out["LCSC Part #"] = out.pop("LCSC_Partnumber")
        return out


def rename_column(rows: list[dict[str, str]], old: str, new: str) -> None:
    for row in rows:
        row[new] = row.pop(old)


def split_designator(designator: str) -> tuple[str, int]:
    match = re.compile(r"(\d+)$").search(designator)
    if match is None:
        return (designator, 0)
    prefix = designator[: match.start()]
    number = int(match.group())
    return (prefix, number)


def make_bom(components: set[Module]):
    bomlines = [line for c in components if (line := _get_bomline(c))]
    bomlines = sorted(
        _compact_bomlines(bomlines),
        key=lambda x: split_designator(x.Designator.split(", ")[0]),
    )

    rows = [line.to_dict() for line in bomlines]
    return rows


def write_bom_jlcpcb(components: set[Module], path: Path) -> None:
    if not path.parent.exists():
        os.makedirs(path.parent)
    with open(path, "w", newline="", encoding="utf-8") as bom_csv:
        rows = make_bom(components)
        if rows:
            writer = csv.DictWriter(
                bom_csv,
                fieldnames=list(rows[0].keys()),
                delimiter=",",
                quotechar='"',
                quoting=csv.QUOTE_MINIMAL,
                lineterminator="\n",
            )
            writer.writeheader()
            writer.writerows(rows)


def _compact_bomlines(bomlines: list[BOMLine]) -> list[BOMLine]:
    compact_bomlines = []
    for row, bomline in enumerate(bomlines):
        # skip PNs that we already added
        if bomline.LCSC_Partnumber in [row.LCSC_Partnumber for row in compact_bomlines]:
            continue

        compact_bomline = bomline
        for other_bomline in bomlines[row + 1 :]:
            if bomline.LCSC_Partnumber == other_bomline.LCSC_Partnumber:
                for key in "Footprint", "Value":
                    if getattr(bomline, key) != getattr(other_bomline, key):
                        logger.warning(
                            f"{key} is not the same for two equal partnumbers "
                            f"{bomline.LCSC_Partnumber}: "
                            f"{bomline.Designator} "
                            f"with {key}: {getattr(bomline, key)} "
                            f"{other_bomline.Designator} "
                            f"with {key}: {getattr(other_bomline, key)}"
                        )
                # Sort designators in bomline by number
                compact_bomline.Designator = ", ".join(
                    sorted(
                        (
                            compact_bomline.Designator + ", " + other_bomline.Designator
                        ).split(", "),
                        key=lambda x: split_designator(x),
                    )
                )
                compact_bomline.Quantity += other_bomline.Quantity
        compact_bomlines += [compact_bomline]

    return compact_bomlines


def _get_bomline(cmp: Module) -> BOMLine | None:
    if not cmp.has_trait(F.has_footprint):
        return
    # TODO make extra trait for this
    if cmp.has_trait(F.has_part_picked) and isinstance(
        cmp.get_trait(F.has_part_picked), F.has_part_removed
    ):
        return

    if not all(
        cmp.has_trait(t)
        for t in (
            F.has_descriptive_properties,
            F.has_designator,
        )
    ):
        logger.warning(f"Missing fields on component '{cmp}'")
        return

    properties = cmp.get_trait(F.has_descriptive_properties).get_properties()
    footprint = cmp.get_trait(F.has_footprint).get_footprint()

    value = (
        cmp.get_trait(F.has_simple_value_representation).get_value()
        if cmp.has_trait(F.has_simple_value_representation)
        else ""
    )
    designator = cmp.get_trait(F.has_designator).get_designator()

    if not footprint.has_trait(F.has_kicad_footprint):
        logger.warning(f"Missing kicad footprint on component '{cmp}'")
        return

    if "LCSC" not in properties:
        return

    manufacturer = (
        properties[DescriptiveProperties.manufacturer]
        if DescriptiveProperties.manufacturer in properties
        else ""
    )
    partnumber = (
        properties[DescriptiveProperties.partno]
        if DescriptiveProperties.partno in properties
        else ""
    )

    footprint_name = footprint.get_trait(
        F.has_kicad_footprint
    ).get_kicad_footprint_name()

    return BOMLine(
        Designator=designator,
        Footprint=footprint_name,
        Quantity=1,
        Value=value,
        Manufacturer=manufacturer,
        Partnumber=partnumber,
        LCSC_Partnumber=properties["LCSC"],
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.parameter import Add, ParameterOperatable
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import Quantity
from faebryk.libs.util import times  # noqa: F401

logger = logging.getLogger(__name__)


class MultiCapacitor(F.Capacitor):
    """
    MultiCapacitor acts a single cap but contains multiple in parallel.
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------

    # ----------------------------------------
    #                 traits
    # ----------------------------------------

    def __init__(self, count: int):
        super().__init__()
        self._count = count

    # Not pickable
    pickable = None

    @L.rt_field
    def capacitors(self) -> list[F.Capacitor]:
        return times(self._count, F.Capacitor)

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------

        self.unnamed[0].connect(*(c.unnamed[0] for c in self.capacitors))
        self.unnamed[1].connect(*(c.unnamed[1] for c in self.capacitors))

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        self.capacitance.alias_is(Add(*(c.capacitance for c in self.capacitors)))
        for c in self.capacitors:
            # TODO use min once available
            self.max_voltage.constrain_le(c.max_voltage)
            self.temperature_coefficient.constrain_superset(c.temperature_coefficient)

    def set_equal_capacitance(self, capacitance: ParameterOperatable):
        op = capacitance / self._count

        self.set_equal_capacitance_each(op)

    def set_equal_capacitance_each(self, capacitance: ParameterOperatable.NumberLike):
        for c in self.capacitors:
            c.capacitance.constrain_subset(capacitance)

    # TODO kinda weird
    def explicit(
        self,
        nominal_capacitance: Quantity | None = None,
        tolerance: float | None = None,
        footprint: str | None = None,
    ):
        for c in self.capacitors:
            c.explicit(nominal_capacitance, tolerance, footprint)

    @classmethod
    def from_capacitors(cls, *capacitors: F.Capacitor):
        # TODO consider merging them more flatly (for multicaps)
        obj = cls(len(capacitors))
        for c_old, c_new in zip(capacitors, obj.capacitors):
            c_new.specialize(c_old)
        return obj



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.exporters.pcb.layout.absolute import LayoutAbsolute
from faebryk.exporters.pcb.layout.extrude import LayoutExtrude
from faebryk.exporters.pcb.layout.typehierarchy import LayoutTypeHierarchy
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class RS485_Bus_Protection(Module):
    """
    RS485 bus protection.
    - Overvoltage protection
    - Overcurrent protection
    - Common mode filter
    - Termination resistor
    - ESD protection
    - Lightning protection

    based on: https://www.mornsun-power.com/public/uploads/pdf/TD(H)541S485H.pdf
    """

    def __init__(self, termination: bool = True, polarization: bool = True) -> None:
        super().__init__()
        self._termination = termination
        self._polarization = polarization

    current_limmiter_resistors = L.list_field(2, F.Resistor)
    gdt: F.GDT
    rs485_tvs: F.ElecSuper_PSM712_ES
    gnd_couple_resistor: F.Resistor
    gnd_couple_capacitor: F.Capacitor
    common_mode_filter: F.Common_Mode_Filter

    power: F.ElectricPower
    rs485_dfp: F.RS485HalfDuplex
    rs485_ufp: F.RS485HalfDuplex
    earth: F.Electrical

    @L.rt_field
    def has_defined_layout(self):
        # PCB layout
        Point = F.has_pcb_position.Point
        L = F.has_pcb_position.layer_type
        LVL = LayoutTypeHierarchy.Level

        self.gnd_couple_resistor.add(
            F.has_pcb_layout_defined(
                layout=LayoutAbsolute(Point((-10.25, -1.5, 90, L.NONE)))
            )
        )

        layouts = [
            LVL(  # GDT
                mod_type=F.GDT,
                layout=LayoutAbsolute(Point((0, 0, 0, L.NONE))),
            ),
            LVL(  # gnd_couple_capacitor
                mod_type=type(self.gnd_couple_capacitor),
                layout=LayoutAbsolute(Point((-7, -1.5, 90, L.NONE))),
            ),
            LVL(  # current_limmiter_resistors
                mod_type=(
                    type(self.current_limmiter_resistors[0]),
                    type(self.current_limmiter_resistors[1]),
                ),
                layout=LayoutExtrude(
                    base=Point((-3, 7.5, 0, L.NONE)),
                    vector=(6, 0, 90),
                    reverse_order=True,
                ),
            ),
            # TODO: fix
            # LVL(  # gnd_couple_resistor
            #    mod_type=type(self.gnd_couple_resistor),
            #    layout=LayoutAbsolute(Point((-10.25, -1.5, 90, L.NONE))),
            # ),
            LVL(  # CompositeTVS
                mod_type=type(self.rs485_tvs),
                layout=LayoutAbsolute(Point((0, 13.5, 0, L.NONE))),
            ),
            LVL(  # CommonModeFilter
                mod_type=F.Common_Mode_Filter,
                layout=LayoutAbsolute(
                    Point((0, 18.5, 90, L.NONE)),
                ),
            ),
        ]
        return F.has_pcb_layout_defined(LayoutTypeHierarchy(layouts))

    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.rs485_dfp, self.rs485_ufp)

    def __preinit__(self):
        if self._termination:
            termination_resistor = self.add(F.Resistor(), name="termination_resistor")
            termination_resistor.resistance.constrain_subset(
                L.Range.from_center_rel(120 * P.ohm, 0.05)
            )
            self.rs485_ufp.diff_pair.p.line.connect_via(
                termination_resistor, self.rs485_ufp.diff_pair.n.line
            )
        if self._polarization:
            polarization_resistors = self.add_to_container(2, F.Resistor)

            polarization_resistors[0].resistance.constrain_subset(
                L.Range(380 * P.ohm, 420 * P.ohm)
            )
            polarization_resistors[1].resistance.constrain_subset(
                L.Range(380 * P.ohm, 420 * P.ohm)
            )
            self.rs485_dfp.diff_pair.p.line.connect_via(
                polarization_resistors[0], self.power.hv
            )
            self.rs485_dfp.diff_pair.n.line.connect_via(
                polarization_resistors[1], self.power.lv
            )

        # ----------------------------------------
        #            parametrization
        # ----------------------------------------
        self.current_limmiter_resistors[0].resistance.constrain_subset(
            L.Range.from_center_rel(2.7 * P.ohm, 0.05)
        )
        self.current_limmiter_resistors[0].max_power.constrain_ge(500 * P.mW)
        self.current_limmiter_resistors[1].resistance.constrain_subset(
            L.Range.from_center_rel(2.7 * P.ohm, 0.05)
        )
        self.current_limmiter_resistors[1].max_power.constrain_ge(500 * P.mW)

        self.gnd_couple_resistor.resistance.constrain_subset(
            L.Range.from_center_rel(1 * P.Mohm, 0.05)
        )
        self.gnd_couple_capacitor.capacitance.constrain_subset(
            L.Range.from_center_rel(1 * P.uF, 0.05)
        )
        self.gnd_couple_capacitor.max_voltage.constrain_ge(2 * P.kV)

        # ----------------------------------------
        #               Connections
        # ----------------------------------------
        # rs485_in/out connections
        self.rs485_dfp.diff_pair.n.line.connect_via(
            [self.common_mode_filter.coil_a, self.current_limmiter_resistors[0]],
            self.rs485_ufp.diff_pair.n.line,
        )
        self.rs485_dfp.diff_pair.p.line.connect_via(
            [self.common_mode_filter.coil_b, self.current_limmiter_resistors[1]],
            self.rs485_ufp.diff_pair.p.line,
        )

        # gdt connections
        self.rs485_ufp.diff_pair.p.line.connect(self.gdt.tube_1)
        self.rs485_ufp.diff_pair.n.line.connect(self.gdt.tube_2)

        # earth connections
        self.power.lv.connect_via(self.gnd_couple_resistor, self.earth)
        self.power.lv.connect_via(self.gnd_couple_capacitor, self.earth)
        self.gdt.common.connect(self.earth)

        # tvs connections
        self.rs485_dfp.connect(self.rs485_tvs.rs485)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from abc import abstractmethod

import faebryk.library._F as F
from faebryk.core.core import Namespace
from faebryk.core.trait import Trait


class LogicOps(Namespace):
    class can_logic(Trait):
        @abstractmethod
        def op(self, *ins: F.Logic) -> F.Logic: ...

    class can_logic_or(can_logic):
        @abstractmethod
        def or_(self, *ins: F.Logic) -> F.Logic: ...

        def op(self, *ins: F.Logic) -> F.Logic:
            return self.or_(*ins)

    class can_logic_and(can_logic):
        @abstractmethod
        def and_(self, *ins: F.Logic) -> F.Logic: ...

        def op(self, *ins: F.Logic) -> F.Logic:
            return self.and_(*ins)

    class can_logic_nor(can_logic):
        @abstractmethod
        def nor(self, *ins: F.Logic) -> F.Logic: ...

        def op(self, *ins: F.Logic) -> F.Logic:
            return self.nor(*ins)

    class can_logic_nand(can_logic):
        @abstractmethod
        def nand(self, *ins: F.Logic) -> F.Logic: ...

        def op(self, *ins: F.Logic) -> F.Logic:
            return self.nand(*ins)

    class can_logic_xor(can_logic):
        @abstractmethod
        def xor(self, *ins: F.Logic) -> F.Logic: ...

        def op(self, *ins: F.Logic) -> F.Logic:
            return self.xor(*ins)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import random
import time
from collections import defaultdict
from contextlib import contextmanager
from enum import Enum, auto
from typing import Callable

from rich.table import Table

from faebryk.libs.logging import rich_to_string
from faebryk.libs.units import P, Quantity, to_si
from faebryk.libs.util import is_numeric_str


class Times:
    class MultiSampleStrategy(Enum):
        AVG = auto()
        ACC = auto()
        MIN = auto()
        MAX = auto()
        MED = auto()
        P80 = auto()
        # Multistrats
        AVG_ACC = auto()
        ALL = auto()

        @property
        def strats(self):
            if self == Times.MultiSampleStrategy.ALL:
                out = list(Times.MultiSampleStrategy)
                out.remove(Times.MultiSampleStrategy.ALL)
                out.remove(Times.MultiSampleStrategy.AVG_ACC)
                return out
            if self == Times.MultiSampleStrategy.AVG_ACC:
                return [Times.MultiSampleStrategy.AVG, Times.MultiSampleStrategy.ACC]
            return [self]

    def __init__(
        self,
        name: str | None = None,
        cnt: int = 1,
        multi_sample_strategy: MultiSampleStrategy = MultiSampleStrategy.ACC,
    ) -> None:
        self.times: dict[str, list[float]] = defaultdict(list)
        self.last_time = time.perf_counter()
        self.name = name

        self.cnt = cnt
        self.strat = multi_sample_strategy

    def add(self, name: str, subcategory: str | None = None):
        if subcategory is not None:
            name = f"{name}:{subcategory}"
        now = time.perf_counter()
        self._add(name, now - self.last_time)
        self.last_time = now

    def _add(self, name: str, val: float):
        self.times[name].append(val)
        if Times._in_measurement:
            if self is Times._in_measurement[0]:
                return
            index = (
                Times._in_measurement.index(self) - 1
                if self in Times._in_measurement
                else -1
            )
            self_name = self.name or hex(id(self))[:4]
            Times._in_measurement[index]._add(f"{self_name}:{name}", val)

    def _format_val(self, val: float, force_unit: str | None = None):
        _val = (val / self.cnt) * P.s
        assert isinstance(_val, Quantity)
        if not force_unit:
            return to_si(_val, "s", num_decimals=2)
        return [f"{_val.m_as(force_unit):.3f}"]

    def get(self, name: str, strat: MultiSampleStrategy | None = None):
        if strat is None:
            strat = self.strat
        if strat == Times.MultiSampleStrategy.ALL:
            raise ValueError("Can't request all samples via get()")
        if strat == Times.MultiSampleStrategy.AVG_ACC:
            raise ValueError("Can't request avg_acc via get()")

        vs = self.times[name]
        match strat:
            case Times.MultiSampleStrategy.AVG:
                v = sum(vs) / len(vs)
            case Times.MultiSampleStrategy.ACC:
                v = sum(vs)
            case Times.MultiSampleStrategy.MIN:
                v = min(vs)
            case Times.MultiSampleStrategy.MAX:
                v = max(vs)
            case Times.MultiSampleStrategy.MED:
                v = sorted(vs)[len(vs) // 2]
            case Times.MultiSampleStrategy.P80:
                v = sorted(vs)[int(0.8 * len(vs))]
        return v

    def get_formatted(self, name: str, strat: MultiSampleStrategy | None = None):
        m, u = self._format_val(self.get(name, strat))
        return f"{m}{u}"

    def __getitem__(self, name: str):
        return self.get(name)

    def to_str(self, force_unit: str | None = None):
        has_multisamples = any(len(vs) > 1 for vs in self.times.values())

        strats = self.strat.strats
        if not has_multisamples:
            strats = [Times.MultiSampleStrategy.AVG]

        table = Table(title="Timings" + (f" (cnt={self.cnt})" if self.cnt > 1 else ""))
        table.add_column("Category", style="cyan")
        if has_multisamples:
            table.add_column("Samples", justify="right")
        for strat in strats:
            c_name = strat.name if has_multisamples else "Value"
            if force_unit:
                c_name = f"{c_name} ({force_unit})"
            table.add_column(
                c_name,
                justify="right",
                style="green",
            )
            if not force_unit:
                table.add_column("Unit", style="yellow")

        rows = []
        seps = []
        for k, vs in self.times.items():
            if k.startswith("_separator"):
                seps.append(len(rows))
                continue
            if any(_k.startswith("_") for _k in k.split(":")):
                continue

            values = []
            for strat in strats:
                v = self.get(k, strat)
                values.extend(self._format_val(v, force_unit=force_unit))

            categories = [k]
            samples = [str(len(vs))] if has_multisamples else []
            rows.append(categories + samples + values)

        # color gradient
        if force_unit:
            for col_i in range(len(rows[0])):
                col = [row[col_i] for row in rows]
                if not all(is_numeric_str(c) for c in col):
                    continue

                # Convert to float for comparison
                values = [float(c) for c in col]
                min_val = min(values)
                max_val = max(values)

                # Skip if all values are the same
                if min_val == max_val:
                    continue

                # Apply color gradient based on percentiles
                # Calculate quartiles
                values_sorted = sorted(values)
                q1_idx = int(len(values_sorted) * 0.25)
                q3_idx = int(len(values_sorted) * 0.75)

                q1 = values_sorted[q1_idx]
                q3 = values_sorted[q3_idx]

                for row_i, val in enumerate(values):
                    # Color based on quartiles: green for below Q1, yellow for Q1-Q3,
                    # red for above Q3
                    if val <= q1:
                        color = "green"
                    elif val <= q3:
                        color = "yellow"
                    else:
                        color = "red"

                    # Apply rich formatting
                    rows[row_i][col_i] = f"[{color}]{rows[row_i][col_i]}[/{color}]"

        for i, row in enumerate(rows):
            if i in seps:
                table.add_section()
            table.add_row(*row)

        return rich_to_string(table)

    def __repr__(self):
        return self.to_str()

    @contextmanager
    def context(self, name: str):
        self.add("_" + name)
        start = time.perf_counter()
        try:
            yield
        finally:
            self._add(name, time.perf_counter() - start)

    _in_measurement: list["Times"] = []

    @contextmanager
    def as_global(self, name: str | None = None, context: bool = False):
        Times._in_measurement.append(self)
        start = time.perf_counter()
        try:
            yield
        finally:
            Times._in_measurement.remove(self)
            if name is not None:
                if context:
                    self._add(name, time.perf_counter() - start)
                else:
                    self.add(name)

    def make_group(self, group_name: str, include_filter: Callable[[str], bool]):
        group = [v for k, vs in self.times.items() if include_filter(k) for v in vs]
        if not group:
            return
        self.times[group_name] = group

    def add_seperator(self):
        self.times["_separator" + hex(random.randint(0, 100000))] = []
